# -*- coding: utf-8 -*-
"""
Created on Thu Mar  2 17:29:41 2017

@author: spauliuk
"""

"""
File ODYM_Classes
Check https://github.com/IndEcol/ODYM for latest version.

Contains class definitions for ODYM

standard abbreviation: msc (material-system-classes)

dependencies:
    numpy >= 1.9
    scipy >= 0.14

Repository for this class, documentation, and tutorials: https://github.com/IndEcol/ODYM

"""
import os
import logging
import numpy as np
import pandas as pd
import xlrd, xlwt

####################################
#      Define classes for ODYM  #
####################################

def __version__():
    return str('1.0') # version number of this file


class Obj(object):
    """
    Class with the object definition for a data object (system, process, flow, ...) in ODYM
    """
    def __init__(self, Name=None, ID=None, UUID=None):
        """ Basic initialisation of Obj."""
        self.Name            = Name # object name
        self.ID              = ID   # object ID
        self.UUID            = UUID # object UUID
        self.Aspects         = {'Time': 'Model time','Cohort': 'Age-cohort','OriginProcess':'Process where flow originates','DestinationProcess':'Destination process of flow','OriginRegion': 'Region where flow originates from','DestinationRegion': 'Region where flow is bound to', 'Good': 'Process, good, or commodity', 'Material': 'Material: ore, alloy, scrap type, ...','Element': 'Chemical element' } # Define the aspects of the system variables
        self.Dimensions      = {'Time': 'Time', 'Process':'Process', 'Region': 'Region', 'Good': 'Process, good, or commodity', 'Material': 'Material: ore, alloy, scrap type, ...','Element': 'Chemical element' } # Define the dimensions of the system variables


        
class Classification(Obj):
  
    """
    Class for aspect classification
    """
    
    def __init__(self, Name = None, ID = None, UUID = None, Dimension = None, Items = None, IDs = None, AdditionalProporties = {}):
        """ Basic initialisation of an item list for alloys, materials, etc."""
        Obj.__init__(self, Name = Name, ID = ID, UUID = UUID) # Hand over parameters to parent class init
        self.Dimension         = Dimension # Dimension of classification: Time, Region, process, material, goods, ...
        self.Items             = Items # list with names of items
        self.IDs               = IDs # list with IDs of items
        self.AdditionalProps   = AdditionalProporties # Like population for regions, element composition for alloys, ...
        
        
class MFAsystem(Obj):

    """
    Class with the definition and methods for a system in ODYM
    """
    
    def __init__(self, Name, Time_Start, Time_End, Geogr_Scope, Unit, IndexTable, Elements, ProcessList = [], FlowDict = {}, StockDict = {}, ParameterDict = {}, Graphical = None, ID = None, UUID = None, ):
        """ Initialisation of MFAsystem."""
        Obj.__init__(self, Name = Name, ID = ID, UUID = UUID) # Hand over parameters to parent class init
        
        self.Time_Start      = Time_Start # start time of model (year: int)
        self.Time_End        = Time_End # end time of model (year: int)
        self.Geogr_Scope     = Geogr_Scope   # geographical boundary (string)
        self.Elements        = Elements # list of chemical elements considered, indicated by atomic numbers
        self.Unit            = Unit    # flow and stock base unit, without 'per yr'
        
        self.ProcessList     = ProcessList # list of processes, processes are referred to by their number
        self.FlowDict        = FlowDict # Dictionary of flows, are indexed by tuples of process they are attached to (p1,p2)
        self.StockDict       = StockDict # Dictionary of stocks, are indexed by process they are located at (p)
        self.ParameterDict   = ParameterDict # Dictionary of of parameters: lifetime, yield rates, etc.
        self.IndexTable      = IndexTable # Dictionary of abbreviations for aspect-classification tuples 
        
        self.Graphical       = Graphical # Dictionary of graphical properties (size in pixel, background color, etc.)

    @property
    def Time_V(self):
        """ Array of all model years"""
        return np.arange(self.Time_Start,self.Time_End +1,1)
        
    @property
    def Time_L(self):
        """ List of all model years"""
        return np.arange(self.Time_Start,self.Time_End +1,1).tolist()
        
    def IndexTableCheck(self):
        """ Check whether chosen classifications fit to dimensions of index table."""
        for indx in self.IndexTable.index:
            if self.IndexTable.loc[indx]['Dimension']  != self.IndexTable.loc[indx]['Classification'].Dimension:
                raise ValueError('Dimension mismatch. Dimension of classifiation needs to fit to dimension of flow or parameter index. Found a mismatch for the following index: {foo}. Check your index table definition!'.format(foo = indx))
        if 'Time' not in self.IndexTable.index:
            raise ValueError(' "Time" aspect must be present in IndexTable. Please check your index table definition!')            
        if 'Element' not in self.IndexTable.index:
            raise ValueError(' "Element" aspect must be present in IndexTable. Please check your index table definition!')                            
        if len(self.IndexTable.loc['Element'].Classification.Items) == 0:
            raise ValueError('Need at least one element in element list, please check your classification definition!')
        if len(self.IndexTable.loc['Time'].Classification.Items) == 0:
            raise ValueError('Need at least one element in Time list, please check your classification definition!')

        return True
           
    def Initialize_FlowValues(self):
        """ This method will construct empty numpy arrays (zeros) for all flows where the value is None and wheree the indices are given."""
        for key in self.FlowDict:
            if self.FlowDict[key].Values is None:
                self.FlowDict[key].Values = np.zeros(tuple([len(self.IndexTable.set_index('IndexLetter').loc[x]['Classification'].Items) for x in self.FlowDict[key].Indices.split(',')]))   
#        Raw code, for development        
#        Indices = 't,Ro,a,e'
#        IndList = Indices.split(',')
#        Dimensions = [len(IndexTable.ix[x]['Classification'].Items) for x in IndList]
#        Values = np.zeros(tuple(Dimensions))       

    def Initialize_StockValues(self):
        """ This method will construct empty numpy arrays (zeros) for all stocks where the value is None and wheree the indices are given."""
        for key in self.StockDict:
            if self.StockDict[key].Values is None:
                self.StockDict[key].Values = np.zeros(tuple([len(self.IndexTable.set_index('IndexLetter').loc[x]['Classification'].Items) for x in self.StockDict[key].Indices.split(',')]))   

    def Initialize_ParameterValues(self):
        """ This method will construct empty numpy arrays (zeros) for all parameters where the value is None and wheree the indices are given."""
        for key in self.ParameterDict:
            if self.ParameterDict[key].Values is None:
                self.ParameterDict[key].Values = np.zeros(tuple([len(self.IndexTable.set_index('IndexLetter').loc[x]['Classification'].Items) for x in self.ParameterDict[key].Indices.split(',')]))                   
                
    def Consistency_Check(self):
        """ Method that check a readily defined system for consistency of dimensions, Value setting, etc. See detailed comments."""                
        
        # 1) Check dimension consistency in index table:
        A = self.IndexTableCheck()
        
        # 2) Check whether all process indices that the flows refer to are in the process list:
        for key in self.FlowDict:
            if self.FlowDict[key].P_Start > len(self.ProcessList) -1:
                raise ValueError('Start process of flow {foo} not present. Check your flow definition!'.format(foo = key))
            if self.FlowDict[key].P_End > len(self.ProcessList) -1:
                raise ValueError('End process of flow {foo} not present. Check your flow definition!'.format(foo = key))                
        
        # 3) Check whethe all flow valua arrays match with the index structure:
        for key in self.FlowDict:
            if tuple([len(self.IndexTable.set_index('IndexLetter').loc[x]['Classification'].Items) for x in self.FlowDict[key].Indices.split(',')]) != self.FlowDict[key].Values.shape:
                raise ValueError('Dimension mismatch. Dimension of flow value array does not fit to flow indices for flow {foo}. Check your flow and flow value definition!'.format(foo = key))
                
        return A, True, True
        
    def Flow_Sum_By_Element(self,FlowKey):
        """ 
        Reduce flow values to a Time x Elements matrix and return as t x e array.
        We take the indices of each flow, e.g., 't,O,D,G,m,e', strip off the ',' to get 'tODGme', 
        add a '->' and the index letters for time and element (here, t and e), 
        and call the Einstein sum function np.einsum with the string 'tODGme->te', 
        and apply it to the flow values. 
        """
        return np.einsum(self.FlowDict[FlowKey].Indices.replace(',','') + '->'+ self.IndexTable.loc['Time'].IndexLetter + self.IndexTable.loc['Element'].IndexLetter ,self.FlowDict[FlowKey].Values) 
    
    def Stock_Sum_By_Element(self,StockKey):
        """ 
        Reduce stock values to a Time x Elements matrix and return as t x e array.
        We take the indices of each stock, e.g., 't,c,G,m,e', strip off the ',' to get 'tcGme', 
        add a '->' and the index letters for time and element (here, t and e), 
        and call the Einstein sum function np.einsum with the string 'tcGme->te', 
        and apply it to the stock values. 
        """
        return np.einsum(self.StockDict[StockKey].Indices.replace(',','') + '->'+ self.IndexTable.loc['Time'].IndexLetter + self.IndexTable.loc['Element'].IndexLetter ,self.StockDict[StockKey].Values) 
                 
    def MassBalance(self, Element = None):
        """ 
        Determines mass balance of MFAsystem
        We take the indices of each flow, e.g., 't,O,D,G,m,e', strip off the ',' to get 'tODGme', 
        add a '->' and the index letters for time and element (here, t and e), 
        and call the Einstein sum function np.einsum with the string 'tODGme->te', 
        and apply it to the flow values. 
        Sum to t and e is subtracted from process where flow is leaving from and added to destination process.
        """
        Bal = np.zeros((len(self.Time_L),len(self.ProcessList),len(self.Elements))) # Balance array: years x process x element: 
        #process position 0 is the balance for the system boundary, the other positions are for the processes, 
        #element position 0 is the balance for the entire mass, the other are for the balance of the individual elements
        
        for key in self.FlowDict: # Add all flows to mass balance
            Bal[:,self.FlowDict[key].P_Start,:] -= self.Flow_Sum_By_Element(key) # Flow leaving a process
            Bal[:,self.FlowDict[key].P_End,:]   += self.Flow_Sum_By_Element(key) # Flow entering a process
            
        for key in self.StockDict: # Add all stock changes to the mass balance
            if  self.StockDict[key].Type == 1:
                Bal[:,self.StockDict[key].P_Res,:] -= self.Stock_Sum_By_Element(key) # 1: net stock change or addition to stock
            elif self.StockDict[key].Type == 2:
                Bal[:,self.StockDict[key].P_Res,:] += self.Stock_Sum_By_Element(key) # 2: removal/release from stock
            
        #add stock changes to process with number 0 ('system boundary, environment of system')
        for key in self.StockDict:
            if  self.StockDict[key].Type == 1:
                Bal[:,0,:] += self.Stock_Sum_By_Element(key) # 1: net stock change or addition to stock
            elif self.StockDict[key].Type == 2:
                Bal[:,0,:] -= self.Stock_Sum_By_Element(key) # 2: removal/release from stock
            
        return Bal
    
    def Check_If_All_Chem_Elements_Are_present(self,FlowKey,AllElementsIndex):
        """
        This method is applicable to systems where the chemical element list contains both 0 ('all' chemical elements) and individual elements.
        It checks whether the sum of the system variable of the other elements equals the entry for element 0.
        This means that the breakdown of the system variable into individual elements has the same mass as the total for all elements.
        AllElementsindex is the position of the element 0 in the element list, typically, it is also 0.
        """
        txe = self.Flow_Sum_By_Element(FlowKey)
        txe_0 = txe[:,AllElementsIndex]
        txe_o = np.delete(txe,AllElementsIndex,axis=1).sum(axis=1)
        if np.allclose(txe_0,txe_o):
            Check = True
        else:
            Check = False
        return Check, txe_0, txe_o  # Check flag, time series for element 'all', time series for all 'other' elements.
        
    def SankeyExport(self,Year, Path, Element): # Export data for given year in excel format for the D3.js Circular Sankey method
        """ Exports MFAsystem to xls Template for the Circular Sankey method."""
        
        TimeIndex = Year - self.Time_Start
        
        myfont = xlwt.Font()
        myfont.bold = True
        mystyle = xlwt.XFStyle()
        mystyle.font = myfont
        
        Result_workbook  = xlwt.Workbook(encoding = 'ascii') 
        Result_worksheet = Result_workbook.add_sheet('Nodes') 
        Result_worksheet.write(0, 0, label = 'Name', style = mystyle)
        Result_worksheet.write(0, 1, label = 'Color', style = mystyle)
        Result_worksheet.write(0, 2, label = 'Orientation', style = mystyle)
        Result_worksheet.write(0, 3, label = 'Width', style = mystyle)
        Result_worksheet.write(0, 4, label = 'Height', style = mystyle)
        Result_worksheet.write(0, 5, label = 'x_position', style = mystyle)
        Result_worksheet.write(0, 6, label = 'y_position', style = mystyle)
        
        for m in range(0,len(self.ProcessList)): 
            if self.ProcessList[m].Graphical is None:
                raise ValueError('Graphical properties of process number {foo} are not set. No export to Sankey possible, as position of process on canvas etc. needs is not specified.'.format(foo = m))
            Result_worksheet.write(m +1, 0, label = self.ProcessList[m].Graphical['Name'])
            Result_worksheet.write(m +1, 1, label = self.ProcessList[m].Graphical['Color'])
            Result_worksheet.write(m +1, 2, label = self.ProcessList[m].Graphical['Angle'])
            Result_worksheet.write(m +1, 3, label = self.ProcessList[m].Graphical['Width'])
            Result_worksheet.write(m +1, 4, label = self.ProcessList[m].Graphical['Height'])
            Result_worksheet.write(m +1, 5, label = self.ProcessList[m].Graphical['xPos'])
            Result_worksheet.write(m +1, 6, label = self.ProcessList[m].Graphical['yPos'])
            
        Result_worksheet = Result_workbook.add_sheet('Flows') 
        Result_worksheet.write(0, 0, label = 'StartNode', style = mystyle)
        Result_worksheet.write(0, 1, label = 'EndNode', style = mystyle)
        Result_worksheet.write(0, 2, label = 'Value', style = mystyle)
        Result_worksheet.write(0, 3, label = 'Color', style = mystyle)
        
        for key in self.FlowDict:
            Result_worksheet.write(m +1, 0, label = self.FlowDict[key].P_Start)
            Result_worksheet.write(m +1, 1, label = self.FlowDict[key].P_End)
            Result_worksheet.write(m +1, 2, label = float(self.Flow_Sum_By_Element(key)[TimeIndex,Element]))
            Result_worksheet.write(m +1, 3, label = self.FlowDict[key].Color)
            
        Result_workbook.save(Path + self.Name + '_' + str(TimeIndex) + '_' + str(Element) + '_Sankey.xls')  
        
        
class Process(Obj):

    """
    Class with the definition and methods for a process in ODYM
    """
    
    def __init__(self, Name = None, ID = None, UUID = None, Bipartite = None, Graphical = None, Extensions = None, Parameters = None):
        """ Basic initialisation of a process."""
        Obj.__init__(self, Name = Name, ID = ID, UUID = UUID) # Hand over parameters to parent class init        
        self.Bipartite = Bipartite   # For bipartite system graphs, a string with value 't' or 'd' for transformation and distribution process indicates which group the process belongs to.
        self.Extensions= Extensions  # Dictionary of     
        self.Graphical = Graphical   # # Dictionary of graphical properties: xPos = None, yPos = None, Orientation = None, Color=None, Width = None, Height=None, 
        
    def add_extension(self,Time = None, Name = None, Value=None, Unit = None, Uncert=None): # Extensions flows that are not part of the system-wide mass balance!
        if self.Extensions is None:
            self.Extensions = []
        self.Extensions.append(Flow(P_Start = self.ID, P_End = None, Time = Time, Name = Name, Unit = Unit, Value = Value, Uncert = Uncert))
      
    def add_parameter(self,Name = None):
        if self.Parameters is None:
            self.Parameters = []
        self.Parameters.append(Parameter(Value = None))
            
class Flow(Obj): # Flow needs to at least have dimension time x element

    """
    Class with the definition and methods for a flow in ODYM
    """
    
    def __init__(self, Name = None, ID = None, UUID = None, P_Start = None, P_End = None, Indices = None, Values=None, Uncert=None, Unit = None, Color = None):
        """ Basic initialisation of a flow."""
        Obj.__init__(self, Name = Name, ID = ID, UUID = UUID) # Hand over parameters to parent class init
        self.P_Start     = P_Start # id of start process of flow (id: int)
        self.P_End       = P_End   # id of end process of flow (id: int)
        self.Indices     = Indices # String with indices as defined in IndexTable, separated by ,: 't,c,p,s,e'
        
        self.Values      = Values   # flow values, np.array, multidimensional, unit is system-wide unit
        self.Uncert      = Uncert  # uncertainty of value in %
        self.Unit        = Unit    # Unit string
        
        self.Color       = Color   # color as string 'R,G,B', where each of R, G, B has a value of 0...255



class Stock(Obj): # Flow needs to at least have dimension time x element

    """
    Class with the definition and methods for a stock in ODYM
    """
    
    def __init__(self, Name = None, ID = None, UUID = None, P_Res = None, Indices = None, Type = None, Values=None, Uncert=None, Unit = None, Color = None):
        """ Basic initialisation of a stock."""
        Obj.__init__(self, Name = Name, ID = ID, UUID = UUID) # Hand over parameters to parent class init
        self.P_Res       = P_Res   # id of process where stock resides (id: int)
        self.Indices     = Indices # String with indices as defined in IndexTable, separated by ,: 't,c,p,s,e'
        self.Type        = Type  # Type is an int value, indicating: 0: stock, 1: (net) stock change or addition to stock, 2: removal from stock
        
        self.Values      = Values  # flow values, np.array, multidimensional, unit is system-wide unit
        self.Uncert      = Uncert  # uncertainty of value in %
        self.Unit        = Unit    # Unit string

        self.Color       = Color   # color as string 'R,G,B', where each of R, G, B has a value of 0...255


class Parameter(Obj):
  
    """
    Class with the definition and methods for parameters
    """
    
    def __init__(self, Name = None, ID = None, UUID = None, P_Res = None, MetaData = None, Indices = None, Values=None, Uncert=None, Unit = None):
        """ Basic initialisation of a parameter."""
        Obj.__init__(self, Name = Name, ID = ID, UUID = UUID) # Hand over parameters to parent class init
        self.P_Res       = P_Res   # id of process to which parameter is assigned (id: int)
        self.Indices     = Indices # String with indices as defined in IndexTable, separated by ,: 't,c,p,s,e'
        self.MetaData    = MetaData # Dictionary with additional metadata
        
        self.Values      = Values   # parameter values, np.array, multidimensional, unit is Unit
        self.Uncert      = Uncert  # uncertainty of value in %
        self.Unit        = Unit   # Unit of parameter values

  
       



        
#   
#
#    


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 11 16:19:39 2014

@authors: Stefan Pauliuk, Uni Freiburg, Germany
"""
import os
import sys
# Put location of 
sys.path.insert(0,  os.path.abspath(os.path.join(os.path.dirname(__file__), '..\\..')) + '\\modules') # add ODYM module directory to system path
#NOTE: Hidden variable __file__ must be know to script for the directory structure to work.
# Therefore: When first using the model, run the entire script with F5 so that the __file__ variable can be created.

import ODYM_Classes as msc # import the ODYM class file
import ODYM_Functions as msf # import the ODYM function file
import numpy as np
import unittest
import pandas as pd


###############################################################################
# Test 1) A simple MFA system with 5 years and 1 chemical element
###############################################################################
# Define results
MyMFA_Test1_Flowshape1 = (5,2) # shape of a flow in system defined below.
F_0_1_Values1          = np.array([[1,1],[0,0],[2,2],[1,1],[4,4]]) # Define some arbitrary inflow data for each year and element
F_1_2_Values1          = np.array([[1,1],[1,1],[32,32],[1,1],[4,4]]) # 
F_2_1_Values1          = np.array([[0,0],[1,1],[30,30],[0,0],[0,0]]) # Define flow back to process 1, make sure mass balance holds.
F_2_0_Values1          = F_0_1_Values1.copy()  # outflow equals inflow as no stocks are present
Bal_Test1              = np.zeros((5,3,2))

# Define dynamic MFA model and fill in values
"""Create Dynamic MFA Models and hand over the pre-defined values."""
# Define a basic classification with 5 years, 1 element, and unity. 'Time' and 'Element' must be present!
ModelClassification_Test1 = {} 
ModelClassification_Test1['Unity']    = msc.Classification(Name = 'Unity',         Dimension = 'Unity',   Items = [1])
ModelClassification_Test1['Time']     = msc.Classification(Name = 'Time',          Dimension = 'Time',    Items = [2010,2011,2012,2013,2014])
ModelClassification_Test1['Element'] = msc.Classification(Name = 'Chem_Elements', Dimension = 'Element', Items = ['All','Fe'])

#extract start and end year
Model_Time_Start = min(ModelClassification_Test1['Time'].Items)
Model_Time_End   = max(ModelClassification_Test1['Time'].Items)

#Define model aspects in right order
Model_Aspects = ['Time','Element','Unity'] # aspects of the model to be considered. Must be the same as the items of the ModelClassification_Test1 dictionary
# Define a simple index table with the three aspects time, element, and unity, using the modelclassification above.
IndexTable = pd.DataFrame({'Aspect'        : Model_Aspects, # 'Time' and 'Element' must be present!
                           'Description'   : ['Time','Element','Unity'],
                           'Dimension'     : ['Time','Element','Unity'], # must be the same as 'Dimenstion' in the classification above
                           'Classification': [ModelClassification_Test1[Aspect] for Aspect in Model_Aspects],
                           'IndexLetter'   : ['t','e','i']}) # Unique one letter (upper or lower case) indices to be used later for calculations. t for time, e for element

IndexTable.set_index('Aspect', inplace = True) # Default indexing of IndexTable, other indices are produced on the fly

# Initialize MFA system
MyMFA_Test1 = msc.MFAsystem(Name = 'TestSystem', 
                      Geogr_Scope = 'World', 
                      Unit = 'Mt', 
                      ProcessList = [], 
                      FlowDict = {}, 
                      StockDict = {},
                      ParameterDict = None, 
                      Time_Start = Model_Time_Start, 
                      Time_End = Model_Time_End, 
                      IndexTable = IndexTable, 
                      Elements = IndexTable.ix['Element'].Classification.Items) 
                      
# Add 2 processes to system
PrL_Name   = ['Environment','Process_1','Process_2']
PrL_Number = [0,1,2]
for m in range(0, len(PrL_Name)):
    MyMFA_Test1.ProcessList.append(msc.Process(Name = PrL_Name[m], ID   = PrL_Number[m]))
    
# Define 4 flows by symbol and names
MyMFA_Test1.FlowDict['F_0_1'] = msc.Flow(Name = 'Input'             , P_Start = 0, P_End = 1, Indices = 't,e', Values=None, Uncert=None)
MyMFA_Test1.FlowDict['F_1_2'] = msc.Flow(Name = 'Processed input'   , P_Start = 1, P_End = 2, Indices = 't,e', Values=None, Uncert=None)
MyMFA_Test1.FlowDict['F_2_1'] = msc.Flow(Name = 'Sent back to 1'    , P_Start = 2, P_End = 1, Indices = 't,e', Values=None, Uncert=None)
MyMFA_Test1.FlowDict['F_2_0'] = msc.Flow(Name = 'Output'            , P_Start = 2, P_End = 0, Indices = 't,e', Values=None, Uncert=None)

MyMFA_Test1.Initialize_FlowValues() # Assign empty arrays to flows according to dimensions specified above

# Assign values to flows, manually: This prodecure mimicks the calculation of results by solving the model equations
MyMFA_Test1.FlowDict['F_0_1'].Values = F_0_1_Values1
MyMFA_Test1.FlowDict['F_1_2'].Values = F_1_2_Values1
MyMFA_Test1.FlowDict['F_2_1'].Values = F_2_1_Values1
MyMFA_Test1.FlowDict['F_2_0'].Values = F_2_0_Values1

Bal = MyMFA_Test1.MassBalance() # Determine Mass Balance    

###############################################################################
# Tests
###############################################################################

class KnownResultsTestCase(unittest.TestCase):

    # For MyMFA_Test1:
    def test_Dimensions_MyMFA_Test1(self):
        """Test Inflow Driven Model with Fixed product lifetime."""
        np.testing.assert_array_equal(MyMFA_Test1.FlowDict['F_1_2'].Values.shape[0],MyMFA_Test1_Flowshape1[0])
        np.testing.assert_array_equal(MyMFA_Test1.FlowDict['F_1_2'].Values.shape[1],MyMFA_Test1_Flowshape1[1])
        np.testing.assert_array_equal(MyMFA_Test1.MassBalance(),Bal_Test1)
           



    if __name__ == '__main__':
        unittest.main()



import os
import sys
import shutil
from importlib.resources import files
from typing import Union
from .runner import run_scenarios


def run_example(path_to_output_dir: Union[str, None] = None,
                remove_existing_output_dir: bool = False):
    """
    Run example scenario and place output to path_to_output_dir.
    If no path_to_output_dir is provided then places results
    to users home directory inside directory "aiphoria_example".
    Existing directory is not deleted and error is shown if
    directory already exists.

    Deleting existing directory can be overridden by setting
    parameter remove_existing_output_dir to True.


    Examples:
    run_example()
    run_example("C:\\results\\aiphoria_example")
    run_example("~/results/aiphoria_example")

    NOTE:
    - Path to output directory MUST BE in absolute format
    (e.g. "C:\\results\\aiphoria_example" (Windows)
    - ~ is expanded to absolute path automatically

    :param path_to_output_dir: Absolute path to output directory
    :param remove_existing_output_dir: If True then removes existing output directory (default = False)
    """
    output_dir_name = "aiphoria_example"
    example_scenario_path = "data/example_scenario.xlsx"

    if path_to_output_dir is None:
        # Place results to users home directory
        path_to_output_dir = os.path.expanduser(os.path.join("~", output_dir_name))
        path_to_output_dir = os.path.realpath(path_to_output_dir)

    if not os.path.isabs(path_to_output_dir):
        sys.stderr.write("ERROR: Path to output directory is not in absolute format\n")
        return False

    if not remove_existing_output_dir and os.path.isdir(path_to_output_dir):
        sys.stderr.write("Directory {} already exists\n".format(path_to_output_dir))
        return False

    # Target directory doesn't exist or is okay to remove
    sys.stdout.write("Using output path = {}\n".format(path_to_output_dir))
    shutil.rmtree(path_to_output_dir, ignore_errors=True)
    os.makedirs(path_to_output_dir, exist_ok=True)

    path_to_example_scenario = files("aiphoria").joinpath(example_scenario_path)
    if not os.path.isfile(path_to_example_scenario):
        sys.stderr.write("ERROR: Example scenario file {} not found, packaging issue\n")
        return False

    current_cwd = os.getcwd()
    os.chdir(path_to_output_dir)
    run_scenarios(path_to_example_scenario,
                  path_to_output_dir=path_to_output_dir,
                  remove_existing_output_dir=remove_existing_output_dir)
    os.chdir(current_cwd)



import os
import json
import zlib
import base64
import webbrowser
from typing import List, Dict, Any
import plotly.graph_objects as go
from PIL import Image
from .datastructures import Scenario, Color
from .parameters import ParameterName
from importlib.resources import files

class DataVisualizer(object):
    def __init__(self):
        pass

    def build_and_show(self, scenarios: List[Scenario],
                       visualizer_params: dict,
                       model_params: dict,
                       combine_to_one_file: bool = True) -> None:
        """
        Build and show the scenarios in the browser.

        :param scenarios: List of Scenario-objects
        :param visualizer_params: Dictionary of visualizer parameters
        :param model_params: Dictionary of model parameters (refer to Builder.py / build_results)
        :param combine_to_one_file: Combine multiple scenarios to one output file (default: False)
        :return: None
        """

        scenario_name_to_info = {}  # Scenario name to info
        scenario_name_to_data = {}  # Scenario name to year to data
        for scenario in scenarios:
            scenario_name_to_info[scenario.name] = self._build_scenario_info(scenario)
            scenario_name_to_data[scenario.name] = self._build_scenario_year_to_data(scenario, visualizer_params)

        if combine_to_one_file:
            # Build combined output file that contains all scenarios
            for scenario in scenarios:
                scenario_name_to_info[scenario.name] = self._build_scenario_info(scenario)
                scenario_name_to_data[scenario.name] = self._build_scenario_year_to_data(scenario, visualizer_params)

            # Generate HTML file for scenarios
            html = self._build_combined_scenario_graph(scenario_name_to_info, scenario_name_to_data, visualizer_params)

            output_path = model_params[ParameterName.OutputPath]
            output_filename = "combined_sankey.html"
            abs_path_to_file = os.path.join(output_path, output_filename)

            with open(abs_path_to_file, "w", encoding="utf-8") as fs:
                fs.write(html)

            if model_params[ParameterName.ShowPlots]:
                webbrowser.open("file://" + os.path.realpath(abs_path_to_file))

        else:
            # Build separate files for each scenario
            for scenario_name in scenario_name_to_data:
                # Generate HTML file for scenario
                scenario_info = scenario_name_to_info[scenario_name]
                scenario_year_to_data = scenario_name_to_data[scenario_name]

                html = self._build_scenario_graph(scenario_name,
                                                  scenario_name_to_info,
                                                  scenario_name_to_data,
                                                  visualizer_params)

                output_path = os.path.join(model_params[ParameterName.OutputPath], scenario_name)
                output_filename = "{}_sankey.html".format(scenario_name)
                abs_path_to_file = os.path.join(output_path, output_filename)

                with open(abs_path_to_file, "w", encoding="utf-8") as fs:
                    fs.write(html)

                if model_params[ParameterName.ShowPlots]:
                    webbrowser.open("file://" + os.path.realpath(abs_path_to_file))

    def _build_scenario_year_to_data(self, scenario: Scenario, params: Dict):
        flow_solver = scenario.flow_solver

        small_node_threshold = params["small_node_threshold"]
        process_transformation_stage_colors = params["process_transformation_stage_colors"]
        virtual_process_graph_labels = params["virtual_process_graph_labels"]
        flow_alpha = params["flow_alpha"]
        virtual_process_color = params["virtual_process_color"]
        virtual_flow_color = params["virtual_flow_color"]

        # Baseline value name and unit names are used for flow data
        baseline_value_name = scenario.model_params[ParameterName.BaselineValueName]
        baseline_unit_name = scenario.model_params[ParameterName.BaselineUnitName]

        # Check if all transformation stages have defined color
        unique_transformation_stages = set()
        year_to_process_to_flows = flow_solver.get_year_to_process_to_flows()
        first_year = list(year_to_process_to_flows.keys())[0]
        for process in year_to_process_to_flows[first_year]:
            unique_transformation_stages.add(process.transformation_stage)

        # Build colors for missing transformation stages or create default color palette
        self._build_default_transformation_stage_colors(unique_transformation_stages,
                                                        process_transformation_stage_colors)

        year_to_data = {}
        year_to_process_to_flows = flow_solver.get_year_to_process_to_flows()
        for year, process_to_flows in year_to_process_to_flows.items():
            year_to_data[year] = {}

            # Per year data
            process_id_to_index = {}
            for index, process in enumerate(process_to_flows):
                process_id_to_index[process.id] = index

            # Per year data of nodes and links for graph
            year_node_labels = []
            year_sources = []
            year_targets = []
            year_node_colors = []
            year_node_positions_x = []
            year_node_positions_y = []
            year_node_custom_data = []
            year_link_values = []
            year_link_colors = []
            year_link_custom_data = []

            for index, process in enumerate(process_to_flows):
                node_label = process.id + "({})".format(process.transformation_stage)
                if process.label_in_graph:
                    node_label = process.label_in_graph

                # Use virtual process color by default
                node_color = virtual_process_color
                if not process.is_virtual:
                    node_color = process_transformation_stage_colors[process.transformation_stage]
                else:
                    # Check if there is a new label for virtual process
                    if process.id in virtual_process_graph_labels:
                        node_label = virtual_process_graph_labels[process.id]

                year_node_labels.append(node_label)
                year_node_colors.append(node_color)
                year_node_positions_x.append(process.position_x)
                year_node_positions_y.append(process.position_y)

                inflows = process_to_flows[process]["in"]
                outflows = process_to_flows[process]["out"]

                # Calculate total inflows and total outflows for Process
                total_inflows = sum([flow.evaluated_value for flow in inflows])
                total_outflows = sum([flow.evaluated_value for flow in outflows])
                for flow in outflows:
                    if flow.source_process_id not in process_id_to_index:
                        print("Source {} not found in process_id_to_index!".format(flow.source_process_id))
                        continue

                    if flow.target_process_id not in process_id_to_index:
                        print("Target {} not found in process_id_to_index!".format(flow.target_process_id))
                        continue

                    source_index = process_id_to_index[flow.source_process_id]
                    target_index = process_id_to_index[flow.target_process_id]
                    year_sources.append(source_index)
                    year_targets.append(target_index)
                    year_link_values.append(flow.evaluated_value)

                    link_color = ""
                    if flow.is_virtual:
                        link_color = virtual_flow_color.lstrip("#")
                        r, g, b = tuple(int(link_color[i:i+2], 16) for i in (0, 2, 4))
                        link_color = "rgba({},{},{},{})".format(r, g, b, flow_alpha)
                    else:
                        link_color = process_transformation_stage_colors[process.transformation_stage]
                        link_color = link_color.lstrip("#")
                        r, g, b = tuple(int(link_color[i:i+2], 16) for i in (0, 2, 4))
                        link_color = "rgba({},{},{},{})".format(r / 255, g / 255, b / 255, flow_alpha)

                    year_link_colors.append(link_color)

                    # Custom data for link
                    year_link_custom_data.append(
                        dict(
                            source_process_id=flow.source_process_id,
                            target_process_id=flow.target_process_id,
                            is_visible=True,
                            is_virtual=flow.is_virtual,
                            evaluated_value=flow.evaluated_value,
                            evaluated_share=flow.evaluated_share,
                            baseline_unit_name=baseline_unit_name,
                            baseline_value_name=baseline_value_name,
                            unit=flow.unit,
                            indicator_names=flow.get_indicator_names(),
                            indicator_units=flow.get_indicator_units(),
                            evaluated_indicator_values=flow.get_all_evaluated_values()
                        )
                    )

                # Custom data for node
                year_node_custom_data.append(
                    dict(
                        node_id=process.id,
                        is_visible=True,
                        is_virtual=process.is_virtual,
                        total_inflows=total_inflows,
                        total_outflows=total_outflows,
                        has_stock=process.stock_lifetime > 0,
                        transformation_stage=process.transformation_stage,
                        stock=dict(
                            distribution_type=process.stock_distribution_type,
                            distribution_params=process.stock_distribution_params,
                            lifetime=process.stock_lifetime,
                        ),
                        x=process.position_x,
                        y=process.position_y,
                    ))

            year_to_data[year] = {
                "labels": year_node_labels,
                "sources": year_sources,
                "targets": year_targets,
                "values": year_link_values,
                "node_colors": year_node_colors,
                "link_colors": year_link_colors,
                "link_custom_data": year_link_custom_data,
                "node_positions_x": year_node_positions_x,
                "node_positions_y": year_node_positions_y,
                "node_custom_data": year_node_custom_data
            }

        # Make list of indicator names
        stock_indicator_names = [baseline_value_name]
        stock_indicator_units = [baseline_unit_name]
        for indicator_name, indicator_entry in scenario.scenario_data.indicator_name_to_indicator.items():
            stock_indicator_names.append(indicator_name)
            stock_indicator_units.append(indicator_entry.unit)

        # Unpack stock data to yearly values
        dsm_baselines = flow_solver.get_baseline_dynamic_stocks()
        dsm_indicators = flow_solver.get_indicator_dynamic_stocks()
        for year_index, year in enumerate(year_to_data.keys()):
            stock_id_to_stock_inflow = {}
            stock_id_to_stock_outflow = {}
            stock_id_to_stock_total = {}
            stock_ids = [stock_id for stock_id in dsm_baselines.keys()]
            for stock_id in stock_ids:
                stock_id_to_stock_inflow[stock_id] = []
                stock_id_to_stock_outflow[stock_id] = []
                stock_id_to_stock_total[stock_id] = []

                process_dsm_baseline = dsm_baselines[stock_id]
                dsm_inflows = process_dsm_baseline.i
                dsm_outflows = process_dsm_baseline.o
                dsm_total = process_dsm_baseline.s
                stock_id_to_stock_inflow[stock_id].append(dsm_inflows[year_index])
                stock_id_to_stock_outflow[stock_id].append(dsm_outflows[year_index])
                stock_id_to_stock_total[stock_id].append(dsm_total[year_index])
                for indicator_name, process_dsm_indicator in dsm_indicators[stock_id].items():
                    dsm_inflows = process_dsm_indicator.i
                    dsm_outflows = process_dsm_indicator.o
                    dsm_total = process_dsm_indicator.s
                    stock_id_to_stock_inflow[stock_id].append(dsm_inflows[year_index])
                    stock_id_to_stock_outflow[stock_id].append(dsm_outflows[year_index])
                    stock_id_to_stock_total[stock_id].append(dsm_total[year_index])

            stock_data = {
                "stock_ids": stock_ids,
                "stock_indicator_names": stock_indicator_names,
                "stock_indicator_units": stock_indicator_units,
                "stock_inflows": stock_id_to_stock_inflow,
                "stock_outflows": stock_id_to_stock_outflow,
                "stock_totals": stock_id_to_stock_total,
            }
            for key in stock_data.keys():
                year_to_data[year][key] = stock_data[key]

        return year_to_data

    def _build_scenario_info(self, scenario: Scenario) -> Dict[str, Any]:
        """
        Build info from Scenario-object.

        :param scenario: Target Scenario-object
        :return: Dictionary (key, value)
        """
        scenario_info = {}
        scenario_info["scenario_name"] = scenario.name
        scenario_info["baseline_value_name"] = scenario.scenario_data.baseline_value_name
        scenario_info["baseline_unit_name"] = scenario.scenario_data.baseline_unit_name
        return scenario_info

    def _build_scenario_graph(self,
                              scenario_name: str,
                              scenario_name_to_info: Dict[str, Any],
                              scenario_name_to_data: Dict[str, Dict] = None,
                              params: Dict = None):

        # Leave only key with target scenario name
        target_scenario_name_to_info = {scenario_name: scenario_name_to_info[scenario_name]}
        target_scenario_name_to_data = {scenario_name: scenario_name_to_data[scenario_name]}

        # Add JS script that is run after the Plotly has loaded
        filename_plotly = os.path.join(os.path.abspath("."), "core", "datavisualizer_data/plotly-3.0.0.min.js")
        filename_pako = os.path.join(os.path.abspath("."), "core", "datavisualizer_data/pako.min.js")
        filename_html = os.path.join(os.path.abspath("."), "core", "datavisualizer_data/datavisualizer_plotly.html")

        # Read HTML file contents
        html = ""
        with open(filename_html, "r", encoding="utf-8") as fs:
            html = fs.read()

        # Read PlotlyJS file contents
        plotly_js = ""
        with open(filename_plotly, "r", encoding="utf-8") as fs:
            plotly_js = fs.read()

        # Read Pako file contents
        pako_js = ""
        with open(filename_pako, "r", encoding="utf-8") as fs:
            pako_js = fs.read()

        # Replace contents with data
        # Plotly
        html = html.replace(
            '<script src="./plotly-3.0.0.min.js"></script>',
            f'<script type="text/javascript">{plotly_js}</script>')

        # Pako
        html = html.replace(
            '<script src="./pako.min.js"></script>',
            f'<script type="text/javascript">{pako_js}</script>')

        # Encode scenario info data as base64-encoded zlib data
        info_json = json.dumps(target_scenario_name_to_info)
        info_compressed = zlib.compress(info_json.encode("utf-8"))
        info_base64 = base64.b64encode(info_compressed).decode("utf-8")
        html = html.replace("// rawScenarioInfo:", "rawScenarioInfo:")
        html = html.replace("{rawScenarioInfo}", json.dumps(info_base64))

        # Encode scenario data as base64-encoded zlib data
        data_json = json.dumps(target_scenario_name_to_data)
        data_compressed = zlib.compress(data_json.encode("utf-8"))
        data_base64 = base64.b64encode(data_compressed).decode("utf-8")
        html = html.replace("// rawScenarioData:", "rawScenarioData:")
        html = html.replace("{rawScenarioData}", json.dumps(data_base64))

        return html

    def _build_combined_scenario_graph(self,
                                       scenario_name_to_info: Dict[str, Any],
                                       scenario_name_to_data: Dict[str, Any] = None,
                                       params: Dict = None):

        # Add JS script that is run after the Plotly has loaded
        filename_plotly = files("aiphoria.core").joinpath("datavisualizer_data/plotly-3.0.0.min.js")
        filename_pako = files("aiphoria.core").joinpath("datavisualizer_data/pako.min.js")
        filename_html = files("aiphoria.core").joinpath("datavisualizer_data/datavisualizer_plotly.html")

        # Read HTML file contents
        html = ""
        with open(filename_html, "r", encoding="utf-8") as fs:
            html = fs.read()

        # Read PlotlyJS file contents
        plotly_js = ""
        with open(filename_plotly, "r", encoding="utf-8") as fs:
            plotly_js = fs.read()

        # Read Pako file contents
        pako_js = ""
        with open(filename_pako, "r", encoding="utf-8") as fs:
            pako_js = fs.read()

        # Replace contents with data
        # Plotly
        html = html.replace(
            '<script src="./plotly-3.0.0.min.js"></script>',
            f'<script type="text/javascript">{plotly_js}</script>')

        # Pako
        html = html.replace(
            '<script src="./pako.min.js"></script>',
            f'<script type="text/javascript">{pako_js}</script>')

        # Encode scenario info data as base64-encoded zlib data
        info_json = json.dumps(scenario_name_to_info)
        info_compressed = zlib.compress(info_json.encode("utf-8"))
        info_base64 = base64.b64encode(info_compressed).decode("utf-8")
        html = html.replace("// rawScenarioInfo:", "rawScenarioInfo:")
        html = html.replace("{rawScenarioInfo}", json.dumps(info_base64))

        # Encode scenario data as base64-encoded zlib data
        data_json = json.dumps(scenario_name_to_data)
        data_compressed = zlib.compress(data_json.encode("utf-8"))
        data_base64 = base64.b64encode(data_compressed).decode("utf-8")
        html = html.replace("// rawScenarioData:", "rawScenarioData:")
        html = html.replace("{rawScenarioData}", json.dumps(data_base64))

        # Metadata
        metadata_json = params["metadata"]
        html = html.replace("// rawMetadata:", "rawMetadata:")
        html = html.replace("{rawMetadata}", json.dumps(metadata_json))

        return html

    def _build_default_transformation_stage_colors(self,
                                                   unique_transformation_stages: set,
                                                   process_transformation_stage_colors: Dict[str, str]):
        """
        Build and fill missing transformation stage colors with default color palette.

        :param unique_transformation_stages: Set of unique transformation stages
        :param process_transformation_stage_colors: Dictionary (transformation stage, Color)
        """

        # Default color palette for 8 transformation stages
        default_color_palette = [
            "#7dda60",
            "#eb5e34",
            "#8c76cf",
            "#5baa11",
            "#3281db",
            "#61b053",
            "#efc3ca",
            "#dfc57b",
        ]

        # Find missing transformation stage names
        defined_transformation_stage_names = set(list(process_transformation_stage_colors.keys()))
        missing_transformation_stage_names = unique_transformation_stages.difference(defined_transformation_stage_names)

        # Fill process_transformation_stage_colors with
        for index, transformation_stage in enumerate(missing_transformation_stage_names):
            color = default_color_palette[index % len(default_color_palette)]
            new_color = Color(params=[transformation_stage, color])
            process_transformation_stage_colors[transformation_stage] = new_color.value



# -*- coding: utf-8 -*-
"""
Class DynamicStockModel
Check https://github.com/IndEcol/ODYM for latest version.

Methods for efficient handling of dynamic stock models (DSMs)

Created on Mon Jun 30 17:21:28 2014

@author: Stefan Pauliuk, NTNU Trondheim, Norway, later Uni Freiburg, Germany
with contributions from
Sebastiaan Deetman, CML, Leiden, NL
Tomer Fishman, IDC Herzliya, IL
Chris Mutel, PSI, Villingen, CH

standard abbreviation: DSM or dsm

dependencies:
    numpy >= 1.9
    scipy >= 0.14

Repository for this class, documentation, and tutorials: https://github.com/IndEcol/ODYM

"""

import numpy as np
import scipy.stats

def __version__():
    """Return a brief version string and statement for this class."""
    return str('1.0'), str('Class DynamicStockModel, dsm. Version 1.0. Last change: July 25th, 2019. Check https://github.com/IndEcol/ODYM for latest version.')


class DynamicStockModel(object):

    """ Class containing a dynamic stock model

    Attributes
    ----------
    t : Series of years or other time intervals
    i : Discrete time series of inflow to stock

    o : Discrete time series of outflow from stock
    o_c :Discrete time series of outflow from stock, by cohort

    s_c : dynamic stock model (stock broken down by year and age- cohort)
    s : Discrete time series for stock, total

    lt : lifetime distribution: dictionary

    pdf: probability density function, distribution of outflow from a specific age-cohort
    
    sf: survival function for different age-cohorts, year x age-cohort table


    name : string, optional
        Name of the dynamic stock model, default is 'DSM'
    """

    """
    Basic initialisation and dimension check methods
    """

    def __init__(self, t=None, i=None, o=None, s=None, lt=None, s_c=None, o_c=None, name='DSM', pdf=None, sf=None):
        """ Init function. Assign the input data to the instance of the object."""
        self.t = t  # optional

        self.i = i  # optional

        self.s = s  # optional
        self.s_c = s_c  # optional

        self.o = o  # optional
        self.o_c = o_c  # optional

        if lt is not None:
            for ThisKey in lt.keys():
                # If we have the same scalar lifetime, stdDev, etc., for all cohorts,
                # replicate this value to full length of the time vector
                if ThisKey != 'Type':
                    if np.array(lt[ThisKey]).shape[0] == 1:
                        lt[ThisKey] = np.tile(lt[ThisKey], len(t))

        self.lt = lt  # optional
        self.name = name  # optional

        self.pdf = pdf # optional
        self.sf  = sf # optional

    """ Part 1: Checks and balances: """

    def dimension_check(self):
        """ This method checks which variables are present and checks whether data types and dimensions match
        """
        # Compile a little report on the presence and dimensions of the elements in the SUT
        try:
            DimReport = str('<br><b> Checking dimensions of dynamic stock model ' + self.name + '.')
            if self.t is not None:
                DimReport += str('Time vector is present with ' + str(len(self.t)) + ' years.<br>')
            else:
                DimReport += str('Time vector is not present.<br>')
            if self.i is not None:
                DimReport += str('Inflow vector is present with ' +
                                 str(len(self.i)) + ' years.<br>')
            else:
                DimReport += str('Inflow is not present.<br>')
            if self.s is not None:
                DimReport += str('Total stock is present with ' + str(len(self.s)) + ' years.<br>')
            else:
                DimReport += str('Total stock is not present.<br>')
            if self.s_c is not None:
                DimReport += str('Stock by cohorts is present with ' + str(len(self.s_c)
                                                                           ) + ' years and ' + str(len(self.s_c[0])) + ' cohorts.<br>')
            else:
                DimReport += str('Stock by cohorts is not present.<br>')
            if self.o is not None:
                DimReport += str('Total outflow is present with ' +
                                 str(len(self.o)) + ' years.<br>')
            else:
                DimReport += str('Total outflow is not present.<br>')
            if self.o_c is not None:
                DimReport += str('Outflow by cohorts is present with ' +
                                 str(len(self.o_c)) + ' years and ' + str(len(self.o_c[0])) + ' cohorts.<br>')
            else:
                DimReport += str('Outflow by cohorts is not present.<br>')
            if self.lt is not None:
                DimReport += str('Lifetime distribution is present with type ' +
                                 str(self.lt['Type']) + '.<br>')
            else:
                DimReport += str('Lifetime distribution is not present.<br>')
            return DimReport
        except:
            return str('<br><b> Checking dimensions of dynamic stock model ' + self.name + ' failed.')

    def compute_stock_change(self):
        """ Determine stock change from time series for stock. Formula: stock_change(t) = stock(t) - stock(t-1)."""
        if self.s is not None:
            stock_change = np.zeros(len(self.s))
            stock_change[0] = self.s[0]
            stock_change[1::] = np.diff(self.s)
            return stock_change
        else:
            return None

    def check_stock_balance(self):
        """ Check wether inflow, outflow, and stock are balanced. If possible, the method returns the vector 'Balance', where Balance = inflow - outflow - stock_change"""
        try:
            Balance = self.i - self.o - self.compute_stock_change()
            return Balance
        except:
            # Could not determine balance. At least one of the variables is not defined.
            return None

    def compute_stock_total(self):
        """Determine total stock as row sum of cohort-specific stock."""
        if self.s is not None:
            return self.s
        else:
            try:
                self.s = self.s_c.sum(axis=1)
                return self.s
            except:
                return None # No stock by cohorts exists, and total stock cannot be computed

    def compute_outflow_total(self):
        """Determine total outflow as row sum of cohort-specific outflow."""
        if self.o is not None:
            # Total outflow is already defined. Doing nothing.
            return self.o
        else:
            try:
                self.o = self.o_c.sum(axis=1)
                return self.o
            except:
                return None # No outflow by cohorts exists, and total outflow cannot be computed
            
    def compute_outflow_mb(self):
        """Compute outflow from process via mass balance. 
           Needed in cases where lifetime is zero."""
        try:
            self.o = self.i - self.compute_stock_change()
            return self.o
        except:
            return None # Variables to compute outflow were not present

    """ Part 2: Lifetime model. """

    def compute_outflow_pdf(self):
        """
        Lifetime model. The method compute outflow_pdf returns an array year-by-cohort of the probability of a item added to stock in year m (aka cohort m) leaves in in year n. This value equals pdf(n,m).
        The pdf is computed from the survival table sf, where the type of the lifetime distribution enters.
        The shape of the output pdf array is NoofYears * NoofYears, but the meaning is years by age-cohorts.
        The method does nothing if the pdf alreay exists.
        """
        if self.pdf is None:
            self.compute_sf() # computation of pdfs moved to this method: compute survival functions sf first, then calculate pdfs from sf.
            self.pdf   = np.zeros((len(self.t), len(self.t)))
            self.pdf[np.diag_indices(len(self.t))] = np.ones(len(self.t)) - self.sf.diagonal(0)
            for m in range(0,len(self.t)):
                self.pdf[np.arange(m+1,len(self.t)),m] = -1 * np.diff(self.sf[np.arange(m,len(self.t)),m])            
            return self.pdf
        else:
            # pdf already exists
            return self.pdf
        
        
    def compute_sf(self): # survival functions
        """
        Survival table self.sf(m,n) denotes the share of an inflow in year n (age-cohort) still present at the end of year m (after m-n years).
        The computation is self.sf(m,n) = ProbDist.sf(m-n), where ProbDist is the appropriate scipy function for the lifetime model chosen.
        For lifetimes 0 the sf is also 0, meaning that the age-cohort leaves during the same year of the inflow.
        The method compute outflow_sf returns an array year-by-cohort of the surviving fraction of a flow added to stock in year m (aka cohort m) in in year n. This value equals sf(n,m).
        This is the only method for the inflow-driven model where the lifetime distribution directly enters the computation. All other stock variables are determined by mass balance.
        The shape of the output sf array is NoofYears * NoofYears, and the meaning is years by age-cohorts.
        The method does nothing if the sf alreay exists. For example, sf could be assigned to the dynamic stock model from an exogenous computation to save time.
        """
        if self.sf is None:
            self.sf = np.zeros((len(self.t), len(self.t)))
            # Perform specific computations and checks for each lifetime distribution:

            if self.lt['Type'] == 'Fixed': # fixed lifetime, age-cohort leaves the stock in the model year when the age specified as 'Mean' is reached.
                for m in range(0, len(self.t)):  # cohort index
                    self.sf[m::,m] = np.multiply(1, (np.arange(0,len(self.t)-m) < self.lt['Mean'][m])) # converts bool to 0/1
                # Example: if Lt is 3.5 years fixed, product will still be there after 0, 1, 2, and 3 years, gone after 4 years.

            # if self.lt['Type'] == 'Simple':  # Implement simple first-order decay
            #     for m in range(0, len(self.t)):  # Loop over each cohort index
            #         mean_lifetime = self.lt['Mean'][m]
            #     # Calculate decay constant k based on mean lifetime
            #         k = ln(2) / half-life #if half-life is provided; otherwise:
            #         k = 1 / mean_lifetime  # For mean lifetime-based decay constant
            #     # Create decay factors for each time step
            #         decay_factors = np.exp(-k * np.arange(0, len(self.t) - m))
            #     # Apply the decay factor to each cohort over time
            #         self.sf[m::, m] = decay_factors

            if self.lt['Type'] == 'Simple':  # Implement simple first-order decay
                for m in range(0, len(self.t)):  # Loop over each cohort index
                    mean_lifetime = self.lt['Mean'][m]

                    # Decide on decay constant k
                    if hasattr(self, 'half_life') and self.half_life is not None:
                        k = np.log(2) / self.half_life
                    else:
                        k = 1 / mean_lifetime  # For mean lifetime-based decay

                    # Create decay factors for each time step
                    decay_factors = np.exp(-k * np.arange(0, len(self.t) - m))

                    # Apply decay factor to the cohort
                    self.sf[m::, m] = decay_factors



            if self.lt['Type'] == 'Normal': # normally distributed lifetime with mean and standard deviation. Watch out for nonzero values 
                # for negative ages, no correction or truncation done here. Cf. note below.
                for m in range(0, len(self.t)):  # cohort index
                    if self.lt['Mean'][m] != 0:  # For products with lifetime of 0, sf == 0
                        self.sf[m::,m] = scipy.stats.norm.sf(np.arange(0,len(self.t)-m), loc=self.lt['Mean'][m], scale=self.lt['StdDev'][m])
                        # NOTE: As normal distributions have nonzero pdf for negative ages, which are physically impossible, 
                        # these outflow contributions can either be ignored (violates the mass balance) or
                        # allocated to the zeroth year of residence, the latter being implemented in the method compute compute_o_c_from_s_c.
                        # As alternative, use lognormal or folded normal distribution options.
                        
            if self.lt['Type'] == 'FoldedNormal': # Folded normal distribution, cf. https://en.wikipedia.org/wiki/Folded_normal_distribution
                for m in range(0, len(self.t)):  # cohort index
                    if self.lt['Mean'][m] != 0:  # For products with lifetime of 0, sf == 0
                        self.sf[m::,m] = scipy.stats.foldnorm.sf(np.arange(0,len(self.t)-m), self.lt['Mean'][m]/self.lt['StdDev'][m], 0, scale=self.lt['StdDev'][m])
                        # NOTE: call this option with the parameters of the normal distribution mu and sigma of curve BEFORE folding,
                        # curve after folding will have different mu and sigma.

            if self.lt['Type'] == 'LogNormal': # lognormal distribution
                # Here, the mean and stddev of the lognormal curve, 
                # not those of the underlying normal distribution, need to be specified! conversion of parameters done here:
                for m in range(0, len(self.t)):  # cohort index
                    if self.lt['Mean'][m] != 0:  # For products with lifetime of 0, sf == 0
                        # calculate parameter mu    of underlying normal distribution:
                        LT_LN = np.log(self.lt['Mean'][m] / np.sqrt(1 + self.lt['Mean'][m] * self.lt['Mean'][m] / (self.lt['StdDev'][m] * self.lt['StdDev'][m]))) 
                        # calculate parameter sigma of underlying normal distribution:
                        SG_LN = np.sqrt(np.log(1 + self.lt['Mean'][m] * self.lt['Mean'][m] / (self.lt['StdDev'][m] * self.lt['StdDev'][m])))
                        # compute survial function
                        self.sf[m::,m] = scipy.stats.lognorm.sf(np.arange(0,len(self.t)-m), s=SG_LN, loc = 0, scale=np.exp(LT_LN)) 
                        # values chosen according to description on
                        # https://docs.scipy.org/doc/scipy-0.13.0/reference/generated/scipy.stats.lognorm.html
                        # Same result as EXCEL function "=LOGNORM.VERT(x;LT_LN;SG_LN;TRUE)"
                        
            if self.lt['Type'] == 'Weibull': # Weibull distribution with standard definition of scale and shape parameters
                for m in range(0, len(self.t)):  # cohort index
                    if self.lt['Shape'][m] != 0:  # For products with lifetime of 0, sf == 0
                        self.sf[m::,m] = scipy.stats.weibull_min.sf(np.arange(0,len(self.t)-m), c=self.lt['Shape'][m], loc = 0, scale=self.lt['Scale'][m])

            elif self.lt['Type'] in ['LandfillDecayWood', 'LandfillDecayPaper']: # FOD method IPCC https://www.ipcc-nggip.iges.or.jp/public/2019rf/pdf/5_Volume5/19R_V5_3_Ch03_SWDS.pdf
                decay_rates = {'LandfillDecayWood': 0.05, 'LandfillDecayPaper': 0.025} # IPCC default decay rates
                doc_values = {'LandfillDecayWood': 0.5, 'LandfillDecayPaper': 0.4} # IPCC default fraction of degradable organic carbon (DOC) for paper and wood
                doc_f_values = {'LandfillDecayWood': 0.77, 'LandfillDecayPaper': 0.5} # IPCC default fraction of DOC that decomposes (DOC_f) for paper and wood

                #The landfill adjustments are scaling factors applied to modify the default decay rates (k values) or half-lives (t1/2) of degradable organic carbon (DOC) in landfills under different conditions.
                landfill_adjustments = {'Dry': 1.0, 'Wet': 1.2, 'Managed': 1.5} # Dry: no adjustment for dry landfills (set to 1); Wet:20% faster decay in wet landfills (k=0.024); Managed:50% faster decay in managed (k=0.03). Note: common k=0.02 is used as the ratios relative to the dry condition

                # Get the value for landfill type
                lt_type = self.lt['Type']
                landfill_type = self.lt["condition"][0]

                k_base = decay_rates[lt_type]
                k_adjusted = k_base * landfill_adjustments.get(landfill_type, 1.0)

                DOC = self.lt.get('DOC', doc_values[lt_type])
                DOC_f = self.lt.get('DOC_f', doc_f_values[lt_type])
                # IPCC default methane correction factor (MCF)
                MCF = self.lt.get('MCF', 1.0)
                # IPCC default fraction of methane in landfill gas (F)
                F = self.lt.get('F', 0.5)
                # IPCC default fraction of methane recovered (R)
                R = self.lt.get('R', 0.0)
                # IPCC default oxidation factor (OX)
                OX = self.lt.get('OX', 0.0)

                for m in range(len(self.t)): # Loop over each cohort index
                    decay_factors = np.exp(-k_adjusted * np.arange(len(self.t) - m)) # Compute decay factors for this cohort
                    DDOCm_decomposed = 1 - decay_factors
                    CH4_generated = DDOCm_decomposed * DOC * DOC_f * MCF * F * (16 / 12)

                    self.sf[m:, m] = decay_factors * DOC * DOC_f * MCF * (1 - R) * (1 - OX)

            return self.sf
        else:
            # sf already exists
            return self.sf
        

    """
    Part 3: Inflow driven model
    Given: inflow, lifetime dist.
    Default order of methods:
    1) determine stock by cohort
    2) determine total stock
    2) determine outflow by cohort
    3) determine total outflow
    4) check mass balance.
    """

    def compute_s_c_inflow_driven(self):
        """ With given inflow and lifetime distribution, the method builds the stock by cohort.
        """
        if self.i is not None:
            if self.lt is not None:
                self.compute_sf()
                self.s_c = np.einsum('c,tc->tc', self.i, self.sf) # See numpy's np.einsum for documentation.
                # This command means: s_c[t,c] = i[c] * sf[t,c] for all t, c
                # from the perspective of the stock the inflow has the dimension age-cohort, 
                # as each inflow(t) is added to the age-cohort c = t
                return self.s_c
            else:
                # No lifetime distribution specified
                return None
        else:
            # No inflow specified
            return None

    def compute_o_c_from_s_c(self):
        """Compute outflow by cohort from stock by cohort."""
        if self.s_c is not None:
            if self.o_c is None:
                self.o_c = np.zeros(self.s_c.shape)
                self.o_c[1::,:] = -1 * np.diff(self.s_c,n=1,axis=0)
                self.o_c[np.diag_indices(len(self.t))] = self.i - np.diag(self.s_c) # allow for outflow in year 0 already
                return self.o_c
            else:
                # o_c already exists. Doing nothing.
                return self.o_c
        else:
            # s_c does not exist. Doing nothing
            return None

    def compute_i_from_s(self, InitialStock):
        """Given a stock at t0 broken down by different cohorts tx ... t0, an "initial stock". 
           This method calculates the original inflow that generated this stock.
           Example: 
        """
        if self.i is None: # only in cases where no inflow has been specified.
            if len(InitialStock) == len(self.t):
                self.i = np.zeros(len(self.t))
                # construct the sf of a product of cohort tc surviving year t 
                # using the lifetime distributions of the past age-cohorts
                self.compute_sf()
                for Cohort in range(0, len(self.t)):
                    if self.sf[-1,Cohort] != 0:
                        self.i[Cohort] = InitialStock[Cohort] / self.sf[-1,Cohort]
                    else:
                        self.i[Cohort] = 0  # Not possible with given lifetime distribution
                return self.i
            else:
                # The length of t and InitialStock needs to be equal
                return None
        else:
            # i already exists. Doing nothing
            return None

    def compute_evolution_initialstock(self,InitialStock,SwitchTime):
        """ Assume InitialStock is a vector that contains the age structure of the stock at time t0, 
        and it covers as many historic cohorts as there are elements in it.
        This method then computes the future stock and outflow from the year SwitchTime onwards.
        Only future years, i.e., years after SwitchTime, are computed.
        NOTE: This method ignores and deletes previously calculated s_c and o_c.
        The InitialStock is a vector of the age-cohort composition of the stock at SwitchTime, with length SwitchTime"""
        if self.lt is not None:
            self.s_c = np.zeros((len(self.t), len(self.t)))
            self.o_c = np.zeros((len(self.t), len(self.t)))
            self.compute_sf()
            # Extract and renormalize array describing fate of initialstock:
            Shares_Left = self.sf[SwitchTime,0:SwitchTime].copy()
            self.s_c[SwitchTime,0:SwitchTime] = InitialStock # Add initial stock to s_c
            self.s_c[SwitchTime::,0:SwitchTime] = np.tile(InitialStock.transpose(),(len(self.t)-SwitchTime,1)) * self.sf[SwitchTime::,0:SwitchTime] / np.tile(Shares_Left,(len(self.t)-SwitchTime,1))
        return self.s_c
    
    

    """
    Part 4: Stock driven model
    Given: total stock, lifetime dist.
    Default order of methods:
    1) determine inflow, outflow by cohort, and stock by cohort
    2) determine total outflow
    3) determine stock change
    4) check mass balance.
    """

    def compute_stock_driven_model(self, NegativeInflowCorrect = False):
        """ With given total stock and lifetime distribution, 
            the method builds the stock by cohort and the inflow.
        """
        if self.s is not None:
            if self.lt is not None:
                self.s_c = np.zeros((len(self.t), len(self.t)))
                self.o_c = np.zeros((len(self.t), len(self.t)))
                self.i = np.zeros(len(self.t))
                # construct the sf of a product of cohort tc remaining in the stock in year t
                self.compute_sf() # Computes sf if not present already.               
                # First year:
                if self.sf[0, 0] != 0: # Else, inflow is 0.
                    self.i[0] = self.s[0] / self.sf[0, 0]
                self.s_c[:, 0] = self.i[0] * self.sf[:, 0] # Future decay of age-cohort of year 0.
                self.o_c[0, 0] = self.i[0] - self.s_c[0, 0]
                # all other years:
                for m in range(1, len(self.t)):  # for all years m, starting in second year
                    # 1) Compute outflow from previous age-cohorts up to m-1
                    self.o_c[m, 0:m] = self.s_c[m-1, 0:m] - self.s_c[m, 0:m] # outflow table is filled row-wise, for each year m.
                    # 2) Determine inflow from mass balance:
                    if NegativeInflowCorrect is False: # if no correction for negative inflows is made 
                        if self.sf[m,m] != 0: # Else, inflow is 0.
                            self.i[m] = (self.s[m] - self.s_c[m, :].sum()) / self.sf[m,m] # allow for outflow during first year by rescaling with 1/sf[m,m]
                        # 3) Add new inflow to stock and determine future decay of new age-cohort
                        self.s_c[m::, m] = self.i[m] * self.sf[m::, m]
                        self.o_c[m, m]   = self.i[m] * (1 - self.sf[m, m])
                    # 2a) Correct remaining stock in cases where inflow would be negative:
                    if NegativeInflowCorrect is True: # if the stock declines faster than according to the lifetime model, this option allows to extract additional stock items.
                        # The negative inflow correction implemented here was developed in a joined effort by Sebastiaan Deetman and Stefan Pauliuk.
                        InflowTest = self.s[m] - self.s_c[m, :].sum()
                        if InflowTest < 0: # if stock-driven model would yield negative inflow
                            Delta = -1 * InflowTest # Delta > 0!
                            self.i[m] = 0 # Set inflow to 0 and distribute mass balance gap onto remaining cohorts:
                            if self.s_c[m,:].sum() != 0:
                                Delta_percent = Delta / self.s_c[m,:].sum() 
                                # Distribute gap equally across all cohorts (each cohort is adjusted by the same %, based on surplus with regards to the prescribed stock)
                                # Delta_percent is a % value <= 100%
                            else:
                                Delta_percent = 0 # stock in this year is already zero, method does not work in this case.
                            # correct for outflow and stock in current and future years
                            # adjust the entire stock AFTER year m as well, stock is lowered in year m, so future cohort survival also needs to decrease.
                            self.o_c[m, :] = self.o_c[m, :] + (self.s_c[m, :] * Delta_percent)  # increase outflow according to the lost fraction of the stock, based on Delta_c
                            self.s_c[m::,0:m] = self.s_c[m::,0:m] * (1-Delta_percent) # shrink future description of stock from previous age-cohorts by factor Delta_percent in current AND future years.
                        else: # If no negative inflow would occur
                            if self.sf[m,m] != 0: # Else, inflow is 0.
                                self.i[m] = (self.s[m] - self.s_c[m, :].sum()) / self.sf[m,m] # allow for outflow during first year by rescaling with 1/sf[m,m]    
                            # Add new inflow to stock and determine future decay of new age-cohort
                            self.s_c[m::, m] = self.i[m] * self.sf[m::, m]
                            self.o_c[m, m]   = self.i[m] * (1 - self.sf[m, m])                                
                        # NOTE: This method of negative inflow correction is only of of many plausible methods of increasing the outflow to keep matching stock levels.
                        # It assumes that the surplus stock is removed in the year that it becomes obsolete. Each cohort loses the same fraction.
                        # Modellers need to try out whether this method leads to justifiable results.
                        # In some situations it is better to change the lifetime assumption than using the NegativeInflowCorrect option.
                    
                return self.s_c, self.o_c, self.i
            else:
                # No lifetime distribution specified
                return None, None, None
        else:
            # No stock specified
            return None, None, None
        

    def compute_stock_driven_model_initialstock(self,InitialStock,SwitchTime,NegativeInflowCorrect = False):
        """ With given total stock and lifetime distribution, the method builds the stock by cohort and the inflow.
        The extra parameter InitialStock is a vector that contains the age structure of the stock at the END of the year Switchtime -1 = t0.
        ***
        Convention 1: Stocks are measured AT THE END OF THE YEAR. Flows occur DURING THE YEAR.
        Convention 2: The model time t spans both historic and future age-cohorts, and the index SwitchTime -1 indicates the first future age-cohort.
        Convention 3: SwitchTime = len(InitialStock) + 1, that means SwitchTime is counted starting from 1 and not 0.
        Convention 4: The future stock time series has 0 as its first len(InitialStock) elements.
        ***
        In the year SwitchTime the model switches from the historic stock to the stock-driven approach. 
        The year SwitchTime is the first year with the stock-driven approach.
        InitialStock contains the age-cohort composition of the stock AT THE END of year SwitchTime -1.
            InitialStock must have length = SwithTime -1.
        For the option "NegativeInflowCorrect", see the explanations for the method compute_stock_driven_model(self, NegativeInflowCorrect = True).
        NegativeInflowCorrect only affects the future stock time series and works exactly as for the stock-driven model without initial stock.
        """
        if self.s is not None:
            if self.lt is not None:
                self.s_c = np.zeros((len(self.t), len(self.t)))
                self.s_c[SwitchTime -2,0:SwitchTime-1] = InitialStock # assign initialstock to stock-by-cohort variable at END OF YEAR SwitchTime (here -1, because indexing starts at 0.).
                self.o_c = np.zeros((len(self.t), len(self.t)))
                self.i = np.zeros(len(self.t))
                
                # construct the sdf of a product of cohort tc leaving the stock in year t
                self.compute_sf() # Computes sf if not present already.  
                # Construct historic inflows
                for c in range(0,SwitchTime -1):
                    if self.sf[SwitchTime -2,c] != 0:
                         self.i[c] = InitialStock[c] / self.sf[SwitchTime -2,c]
                    else:
                         self.i[c] = InitialStock[c]
                         
                # Add stock from historic inflow
                self.s_c[:,0:SwitchTime-1] = np.einsum('tc,c->tc',self.sf[:,0:SwitchTime-1],self.i[0:SwitchTime-1])
                # calculate historic outflow
                for m in range(0,SwitchTime-1):
                    self.o_c[m, m]    = self.i[m] * (1 - self.sf[m, m])
                    self.o_c[m+1::,m] = self.s_c[m:-1,m] - self.s_c[m+1::,m]
                # for future: year-by-year computation, starting from SwitchTime
                if NegativeInflowCorrect is False:
                    for m in range(SwitchTime-1, len(self.t)):  # for all years m, starting at SwitchTime
                        # 1) Determine inflow from mass balance:
                        if self.sf[m,m] != 0: # Else, inflow is 0.
                            self.i[m] = (self.s[m] - self.s_c[m, :].sum()) / self.sf[m,m] # allow for outflow during first year by rescaling with 1/sf[m,m]
                        # NOTE: The stock-driven method may lead to negative inflows, if the stock development is in contradiction with the lifetime model.
                        # In such situations the lifetime assumption must be changed, either by directly using different lifetime values or by adjusting the outlfows, 
                        # cf. the option NegativeInflowCorrect in the method compute_stock_driven_model.
                        # 2) Add new inflow to stock and determine future decay of new age-cohort
                        self.s_c[m::, m]  = self.i[m] * self.sf[m::, m]
                        self.o_c[m, m]    = self.i[m] * (1 - self.sf[m, m])
                        self.o_c[m+1::,m] = self.s_c[m:-1,m] - self.s_c[m+1::,m]
                if NegativeInflowCorrect is True:
                    for m in range(SwitchTime-1, len(self.t)):  # for all years m, starting at SwitchTime
                        self.o_c[m, 0:m] = self.s_c[m-1, 0:m] - self.s_c[m, 0:m] # outflow table is filled row-wise, for each year m.
                        # 1) Determine text inflow from mass balance:
                        InflowTest = self.s[m] - self.s_c[m, :].sum()
                        if InflowTest < 0: 
                            Delta = -1 * InflowTest # Delta > 0!
                            self.i[m] = 0 # Set inflow to 0 and distribute mass balance gap onto remaining cohorts:
                            if self.s_c[m,:].sum() != 0:
                                Delta_percent = Delta / self.s_c[m,:].sum() 
                                # Distribute gap equally across all cohorts (each cohort is adjusted by the same %, based on surplus with regards to the prescribed stock)
                                # Delta_percent is a % value <= 100%
                            else:
                                Delta_percent = 0 # stock in this year is already zero, method does not work in this case.
                            # correct for outflow and stock in current and future years
                            # adjust the entire stock AFTER year m as well, stock is lowered in year m, so future cohort survival also needs to decrease.
                            # print(InflowTest)
                            # print((self.s_c[m, :] * Delta_percent).sum())
                            # print('_')
                            self.o_c[m, :] = self.o_c[m, :] + (self.s_c[m, :] * Delta_percent).copy()  # increase outflow according to the lost fraction of the stock, based on Delta_c
                            self.s_c[m::,0:m] = self.s_c[m::,0:m] * (1-Delta_percent.copy()) # shrink future description of stock from previous age-cohorts by factor Delta_percent in current AND future years.
                        else:
                            if self.sf[m,m] != 0: # Else, inflow is 0.
                                self.i[m] = (self.s[m] - self.s_c[m, :].sum()) / self.sf[m,m] # allow for outflow during first year by rescaling with 1/sf[m,m]
                            # 2) Add new inflow to stock and determine future decay of new age-cohort
                            self.s_c[m::, m]  = self.i[m] * self.sf[m::, m]
                            self.o_c[m, m]    = self.i[m] * (1 - self.sf[m, m])
                # Add historic stock series to total stock s:
                self.s[0:SwitchTime-1]= self.s_c[0:SwitchTime-1,:].sum(axis =1).copy()                    
                return self.s_c, self.o_c, self.i
            else:
                # No lifetime distribution specified
                return None, None, None
        else:
            # No stock specified
            return None, None, None       
        
  
    def compute_stock_driven_model_initialstock_typesplit(self,FutureStock,InitialStock,SFArrayCombined,TypeSplit):
        """ 
        With given total future stock and lifetime distribution, the method builds the stock by cohort and the inflow.
        The age structure of the initial stock is given for each technology, and a type split of total inflow into different technology types is given as well.
        
        SPECIFICATION: Stocks are always measured AT THE END of the discrete time interval.
        
        Indices:
          t: time: Entire time frame: from earliest age-cohort to latest model year.
          c: age-cohort: same as time.
          T: Switch time: DEFINED as first year where historic stock is NOT present, = last year where historic stock is present +1.
             Switchtime is calculated internally, by subtracting the length of the historic stock from the total model length.
          g: product type
        
        Data:
          FutureStock[t],           total future stock at end of each year, starting at T
          InitialStock[c,g],        0...T-1;0...T-1, stock at the end of T-1, by age-cohort c, ranging from 0...T-1, and product type g
                                    c-dimension has full length, all future years must be 0.
          SFArrayCombined[t,c,g],   Survival function of age-cohort c at end of year t for product type g
                                    this array spans both historic and future age-cohorts
          Typesplit[t,g],           splits total inflow into product types for future years 
            
        The extra parameter InitialStock is a vector that contains the age structure of the stock at time t0, and it covers as many historic cohorts as there are elements in it.
        In the year SwitchTime the model switches from the historic stock to the stock-driven approach.
        Only future years, i.e., years after SwitchTime, are computed and returned.
        The InitialStock is a vector of the age-cohort composition of the stock at SwitchTime, with length SwitchTime.
        The parameter TypeSplit splits the total inflow into Ng types. """
        
        if self.s is not None:
            if self.lt is not None:
                
                SwitchTime = SFArrayCombined.shape[0] - FutureStock.shape[0]
                Ntt        = SFArrayCombined.shape[0] # Total no of years
                Nt0        = FutureStock.shape[0]     # No of future years
                Ng         = SFArrayCombined.shape[2] # No of product groups
                
                s_cg = np.zeros((Nt0,Ntt,Ng)) # stock for future years, all age-cohorts and product
                o_cg = np.zeros((Nt0,Ntt,Ng)) # outflow by future years, all cohorts and products
                i_g  = np.zeros((Ntt,Ng))     # inflow by product
                
                # Construct historic inflows
                for c in range(0,SwitchTime): # for all historic age-cohorts til SwitchTime - 1:
                    for g in range(0,Ng):
                        if SFArrayCombined[SwitchTime-1,c,g] != 0:
                         i_g[c,g] = InitialStock[c,g] / SFArrayCombined[SwitchTime-1,c,g]
                         
                         # if InitialStock is 0, historic inflow also remains 0, 
                         # as it has no impact on future anymore.
                         
                         # If survival function is 0 but initial stock is not, the data are inconsisent and need to be revised.
                         # For example, a safety-relevant device with 5 years fixed lifetime but a 10 year old device is present.
                         # Such items will be ignored and break the mass balance.
            
                # year-by-year computation, starting from SwitchTime
                for t in range(SwitchTime, Ntt):  # for all years t, starting at SwitchTime
                    # 1) Compute stock at the end of the year:
                    s_cg[t - SwitchTime,:,:] = np.einsum('cg,cg->cg',i_g,SFArrayCombined[t,:,:])
                    # 2) Compute outflow during year t from previous age-cohorts:
                    if t == SwitchTime:
                        o_cg[t -SwitchTime,:,:] = InitialStock - s_cg[t -SwitchTime,:,:]
                    else:
                        o_cg[t -SwitchTime,:,:] = s_cg[t -SwitchTime -1,:,:] - s_cg[t -SwitchTime,:,:] # outflow table is filled row-wise, for each year t.
                    # 3) Determine total inflow from mass balance:
                    i0 = FutureStock[t -SwitchTime] - s_cg[t - SwitchTime,:,:].sum()
                    # 4) Add new inflow to stock and determine future decay of new age-cohort
                    i_g[t,:] = TypeSplit[t -SwitchTime,:] * i0
                    for g in range(0,Ng): # Correct for share of inflow leaving during first year.
                        if SFArrayCombined[t,t,g] != 0: # Else, inflow leaves within the same year and stock modelling is useless
                            i_g[t,g] = i_g[t,g] / SFArrayCombined[t,t,g] # allow for outflow during first year by rescaling with 1/SF[t,t,g]
                        s_cg[t -SwitchTime,t,g]  = i_g[t,g] * SFArrayCombined[t,t,g]
                        o_cg[t -SwitchTime,t,g]  = i_g[t,g] * (1 - SFArrayCombined[t,t,g])
                    
                # Add total values of parameter to enable mass balance check:
                self.s_c = s_cg.sum(axis =2)
                self.o_c = o_cg.sum(axis =2)
                self.i   =  i_g[SwitchTime::,:].sum(axis =1)
                
                return s_cg, o_cg, i_g
            else:
                # No lifetime distribution specified
                return None, None, None
        else:
            # No stock specified
            return None, None, None      
        
    def compute_stock_driven_model_initialstock_typesplit_negativeinflowcorrect(self,SwitchTime,InitialStock,SFArrayCombined,TypeSplit,NegativeInflowCorrect = False):
        """ 
        With given total future stock and lifetime distribution, the method builds the stock by cohort and the inflow.
        The age structure of the initial stock is given for each technology, and a type split of total inflow into different technology types is given as well.
        For the option "NegativeInflowCorrect", see the explanations for the method compute_stock_driven_model(self, NegativeInflowCorrect = True).
        NegativeInflowCorrect only affects the future stock time series and works exactly as for the stock-driven model without initial stock.
        
        SPECIFICATION: Stocks are always measured AT THE END of the discrete time interval.
        
        Indices:
          t: time: Entire time frame: from earliest age-cohort to latest model year.
          c: age-cohort: same as time.
          T: Switch time: DEFINED as first year where historic stock is NOT present, = last year where historic stock is present +1.
             Switchtime must be given as argument. Example: if the first three age-cohorts are historic, SwitchTime is 3, which indicates the 4th year.
             That also means that the first 3 time-entries for the stock and typesplit arrays must be 0.
          g: product type
        
        Data:
          s[t],                     total future stock time series, at end of each year, starting at T, trailing 0s for historic years.
                                    ! is not handed over with the function call but earlier, when defining the dsm.
          InitialStock[c,g],        0...T-1;0...T-1, stock at the end of T-1, by age-cohort c, ranging from 0...T-1, and product type g
                                    c-dimension has full length, all future years must be 0.
          SFArrayCombined[t,c,g],   Survival function of age-cohort c at end of year t for product type g
                                    this array spans both historic and future age-cohorts
          Typesplit[t,g],           splits total inflow into product types for future years 
          NegativeInflowCorrect     BOOL, retains items in stock if their leaving would lead to negative inflows. 
            
        The extra parameter InitialStock is a vector that contains the age structure of the stock at time t0, and it covers as many historic cohorts as there are elements in it.
        In the year SwitchTime the model switches from the historic stock to the stock-driven approach.
        Only future years, i.e., years after SwitchTime, are computed and returned.
        The InitialStock is a vector of the age-cohort composition of the stock at SwitchTime, with length SwitchTime.
        The parameter TypeSplit splits the total inflow into Ng types. """
        
        if self.s is not None:
            if self.lt is not None:
                
                Ntt        = SFArrayCombined.shape[0] # Total no of years
                Ng         = SFArrayCombined.shape[2] # No of product groups
                
                s_cg = np.zeros((Ntt,Ntt,Ng)) # stock for future years, all age-cohorts and products
                o_cg = np.zeros((Ntt,Ntt,Ng)) # outflow by future years, all cohorts and products
                i_g  = np.zeros((Ntt,Ng))     # inflow for all years by product
                NIC_Flags = np.zeros((Ntt,1)) # inflow flog for future years, will be set to calculated negative inflow value if negative inflow occurs and is corrected for.
                
                self.s_c = np.zeros((len(self.t), len(self.t)))
                self.o_c = np.zeros((len(self.t), len(self.t)))
                self.i   = np.zeros(len(self.t))
                
                # construct the sdf of a product of cohort tc leaving the stock in year t
                self.compute_sf() # Computes sf if not present already.
                # Construct historic inflows
                for c in range(0,SwitchTime): # for all historic age-cohorts til SwitchTime - 1:
                    for g in range(0,Ng):
                        if SFArrayCombined[SwitchTime-1,c,g] != 0:
                            i_g[c,g] = InitialStock[c,g] / SFArrayCombined[SwitchTime-1,c,g]
                         
                         # if InitialStock is 0, historic inflow also remains 0, 
                         # as it has no impact on future anymore.
                         
                         # If survival function is 0 but initial stock is not, the data are inconsisent and need to be revised.
                         # For example, a safety-relevant device with 5 years fixed lifetime but a 10 year old device is present.
                         # Such items will be ignored and break the mass balance.
                         
                # Compute stocks from historic inflows
                s_cg[:,0:SwitchTime,:] = np.einsum('tcg,cg->tcg',SFArrayCombined[:,0:SwitchTime,:],i_g[0:SwitchTime,:])
                # calculate historic outflows
                for m in range(0,SwitchTime):
                    o_cg[m,m,:]      = i_g[m,:] * (1 - SFArrayCombined[m,m,:])
                    o_cg[m+1::,m,:]  = s_cg[m:-1,m,:] - s_cg[m+1::,m,:]
                # add historic age-cohorts to total stock:
                self.s[0:SwitchTime] = np.einsum('tcg->t',s_cg[0:SwitchTime,:,:])
                
                # for future: year-by-year computation, starting from SwitchTime
                if NegativeInflowCorrect is False:
                    for m in range(SwitchTime, len(self.t)):  # for all years m, starting at SwitchTime
                        # 1) Determine inflow from mass balance:
                        i0_test = self.s[m] - s_cg[m,:,:].sum()
                        if i0_test < 0:
                            NIC_Flags[m] = i0_test
                        for g in range(0,Ng):
                            if SFArrayCombined[m,m,g] != 0: # Else, inflow is 0.
                                i_g[m,g] = TypeSplit[m,g] * i0_test / SFArrayCombined[m,m,g] # allow for outflow during first year by rescaling with 1/sf[m,m]
                                # NOTE: The stock-driven method may lead to negative inflows, if the stock development is in contradiction with the lifetime model.
                                # In such situations the lifetime assumption must be changed, either by directly using different lifetime values or by adjusting the outlfows, 
                                # cf. the option NegativeInflowCorrect in the method compute_stock_driven_model.
                                # 2) Add new inflow to stock and determine future decay of new age-cohort
                            s_cg[m::,m,g]   = i_g[m,g] * SFArrayCombined[m::,m,g]
                            o_cg[m,m,g]     = i_g[m,g] * (1 - SFArrayCombined[m,m,g])
                            o_cg[m+1::,m,g] = s_cg[m:-1,m,g] - s_cg[m+1::,m,g]
                            
                if NegativeInflowCorrect is True:
                    for m in range(SwitchTime, len(self.t)):  # for all years m, starting at SwitchTime
                        # 1) Determine inflow from mass balance:
                        i0_test = self.s[m] - s_cg[m,:,:].sum()
                        if i0_test < 0:
                            NIC_Flags[m] = i0_test                        
                            Delta = -1 * i0_test # Delta > 0!
                            i_g[m,:] = 0 # Set inflow to 0 and distribute mass balance gap onto remaining cohorts:
                            if s_cg[m,:,:].sum() != 0:
                                Delta_percent = Delta / s_cg[m,:,:].sum() 
                                # Distribute gap equally across all cohorts (each cohort is adjusted by the same %, based on surplus with regards to the prescribed stock)
                                # Delta_percent is a % value <= 100%
                            else:
                                Delta_percent = 0 # stock in this year is already zero, method does not work in this case.
                            # correct for outflow and stock in current and future years
                            # adjust the entire stock AFTER year m as well, stock is lowered in year m, so future cohort survival also needs to decrease.
                            o_cg[m, :,:]    = o_cg[m, :,:]    + (s_cg[m, :,:] * Delta_percent).copy()  # increase outflow according to the lost fraction of the stock, based on Delta_c
                            s_cg[m::,0:m,:] = s_cg[m::,0:m,:] * (1-Delta_percent.copy())               # shrink future description of stock from previous age-cohorts by factor Delta_percent in current AND future years.
                            o_cg[m+1::,:,:] = s_cg[m:-1,:,:] - s_cg[m+1::,:,:]                         # recalculate future outflows
                        
                        else:       
                            for g in range(0,Ng):
                                if SFArrayCombined[m,m,g] != 0: # Else, inflow is 0.
                                    i_g[m,g] = TypeSplit[m,g] * i0_test / SFArrayCombined[m,m,g] # allow for outflow during first year by rescaling with 1/sf[m,m]
                                    # NOTE: The stock-driven method may lead to negative inflows, if the stock development is in contradiction with the lifetime model.
                                    # In such situations the lifetime assumption must be changed, either by directly using different lifetime values or by adjusting the outlfows, 
                                    # cf. the option NegativeInflowCorrect in the method compute_stock_driven_model.
                                    # 2) Add new inflow to stock and determine future decay of new age-cohort
                                s_cg[m::,m,g]   = i_g[m,g] * SFArrayCombined[m::,m,g]
                                o_cg[m,m,g]     = i_g[m,g] * (1 - SFArrayCombined[m,m,g])
                                o_cg[m+1::,m,g] = s_cg[m:-1,m,g] - s_cg[m+1::,m,g]    
                                
                # Add total values of parameter to enable mass balance check:
                self.s_c = s_cg.sum(axis =2)
                self.o_c = o_cg.sum(axis =2)
                self.i   = i_g.sum(axis =1)
                
                return s_cg, o_cg, i_g, NIC_Flags
            
            else:
                # No lifetime distribution specified
                return None, None, None, None
        else:
            # No stock specified
            return None, None, None, None
      
        

#
#
# The end.
#




import copy
import sys
from typing import List, Dict, Tuple, Union
import numpy as np
import pandas as pd
import tqdm as tqdm
from pandas import DataFrame

from .types import FunctionType
from .datastructures import Process, Flow, Stock, ScenarioData, Scenario, Indicator
# from .flowmodifiersolver import FlowModifierSolver
from .parameters import ParameterName, StockDistributionType, StockDistributionParameter, ParameterScenarioType
from aiphoria.lib.odym.modules.dynamic_stock_model import DynamicStockModel


# Solves flows to absolute values
class FlowSolver(object):
    """
    Solves flows to absolute values, evaluates Process in- and outflow values,
    and handles dynamic stocks.
    """
    _virtual_process_id_prefix = "VP_"
    _virtual_flow_id_prefix = "VF_"
    _max_iterations = 100000
    _virtual_process_transformation_stage = "Virtual"

    def __init__(self, scenario: Scenario = None, reset_evaluated_values: bool = True):
        self._reset_evaluated_values = reset_evaluated_values
        self._scenario = scenario

        # Prioritized transformation stages
        self._model_params = self._scenario.model_params
        self._prioritized_locations = self._model_params[ParameterName.PrioritizeLocations]
        self._prioritized_transformation_stages = self._model_params[ParameterName.PrioritizeTransformationStages]

        # Time
        self._year_start = self._scenario.scenario_data.start_year
        self._year_end = self._scenario.scenario_data.end_year
        self._years = self._scenario.scenario_data.years
        self._year_current = self._year_start
        self._year_prev = self._year_current

        # Year to Process/Flow/Flow IDs mappings
        self._year_to_process_id_to_process = scenario.scenario_data.year_to_process_id_to_process
        self._year_to_process_id_to_flow_ids = scenario.scenario_data.year_to_process_id_to_flow_ids
        self._year_to_flow_id_to_flow = scenario.scenario_data.year_to_flow_id_to_flow

        # Stocks
        self._all_stocks = self._scenario.scenario_data.stocks
        self._process_id_to_stock = self._scenario.scenario_data.process_id_to_stock

        # Unique Process IDs and Flow IDs to Process/Flow
        self._unique_process_id_to_process = self._scenario.scenario_data.unique_process_id_to_process
        self._unique_flow_id_to_flow = self._scenario.scenario_data.unique_flow_id_to_flow

        # Virtual flows
        self._use_virtual_flows = self._scenario.scenario_data.use_virtual_flows
        self._virtual_flows_epsilon = self._scenario.scenario_data.virtual_flows_epsilon

        # Current timestep data
        self._current_process_id_to_process = self._year_to_process_id_to_process[self._year_current]
        self._current_process_id_to_flow_ids = self._year_to_process_id_to_flow_ids[self._year_current]
        self._current_flow_id_to_flow = self._year_to_flow_id_to_flow[self._year_current]

        # Prepare flows for all timesteps
        for year, flow_id_to_flow in self._year_to_flow_id_to_flow.items():
            self._prepare_flows_for_timestep(flow_id_to_flow, year)

        # Get and store indicator names from scenario.scenario_data
        self._indicator_name_to_indicator = scenario.scenario_data.indicator_name_to_indicator
        self._indicators = {name: indicator for name, indicator in self._indicator_name_to_indicator.items()}

        # Baseline indicator name (e.g. Solid wood equivalent) and unit name (e.g. 'Mm3')
        self._baseline_value_name = self._scenario.scenario_data.baseline_value_name
        self._baseline_unit_name = self._scenario.scenario_data.baseline_unit_name

        # Stock ID -> Baseline value DSM
        self._stock_id_to_baseline_dsm = {}

        # Stock ID -> Indicator name -> DSM
        self._stock_id_to_indicator_name_to_dsm = {}

    def get_scenario(self) -> Scenario:
        """
        Get Scenario that FlowSolver is using.

        :return: Scenario-object
        """
        return self._scenario

    def get_all_stocks(self) -> List[Stock]:
        """
        Get list of all Stocks.

        :return: List of all Stocks
        """
        return self._all_stocks

    def get_unique_processes(self) -> Dict[str, Process]:
        """
        Get dictionary of all unique Process ID to Process.

        :return: Dictionary (Process ID (str) -> Process)
        """
        return self._unique_process_id_to_process

    def get_unique_flows(self) -> Dict[str, Flow]:
        """
        Get dictionary of all unique Flow ID to Flow.

        :return: Dictionary (Flow ID (str) -> Flow)
        """
        return self._unique_flow_id_to_flow

    # Utility methods
    def get_processes_as_dataframe(self) -> DataFrame:
        """
        Get Process information as DataFrame for every year:
            - Process ID
            - Total inflows (baseline)
            - Total outflows (baseline)
            - Total inflows (indicator N)
            - Total outflows (indicator N)
            - Total inflows (indicator N+1)
            - Total outflows (indicator N+1)
            - ...

        :return: DataFrame
        """
        col_names = ["Year", "Process ID"]
        col_names += ["Total inflows, {} ({})".format(self._baseline_value_name, self._baseline_unit_name)]
        col_names += ["Total outflows, {} ({})".format(self._baseline_value_name, self._baseline_unit_name)]
        for indicator in self.get_indicator_name_to_indicator().values():
            col_names += ["Total inflows, {} ({})".format(indicator.name, indicator.unit)]
            col_names += ["Total outflows, {} ({})".format(indicator.name, indicator.unit)]

        rows = []
        for year, process_id_to_process in self._year_to_process_id_to_process.items():
            for process_id, process in process_id_to_process.items():
                new_row = [year, process_id]
                new_row += [self.get_process_inflows_total(process_id, year)]
                new_row += [self.get_process_outflows_total(process_id, year)]
                for indicator in self.get_indicator_name_to_indicator().values():
                    new_row += [self._get_process_indicator_inflows_total(process_id, indicator.name, year)]
                    new_row += [self._get_process_indicator_outflows_total(process_id, indicator.name, year)]
                rows.append(new_row)

        df = pd.DataFrame(rows, columns=col_names)
        return df

    def get_flows_as_dataframe(self) -> DataFrame:
        """
        Get all Flow information for all years in DataFrame:
            - Flow ID
            - Source Process ID
            - Target Process ID
            - Baseline value (baseline unit)
            - Indicator N value (indicator N unit)
            - Indicator N+1 value (indicator N+1 unit)
            - ...

        :return: DataFrame
        """
        col_names = ["Year", "Flow ID", "Source Process ID", "Target Process ID", "Flow share"]
        col_names += ["{} ({})".format(self._baseline_value_name, self._baseline_unit_name)]
        col_names += ["{} ({})".format(ind.name, ind.unit) for ind in self.get_indicator_name_to_indicator().values()]

        rows = []
        for year, flow_id_to_flow in self._year_to_flow_id_to_flow.items():
            for flow_id, flow in flow_id_to_flow.items():
                if not isinstance(flow, Flow):
                    continue

                new_row = [year, flow_id, flow.source_process_id, flow.target_process_id, flow.evaluated_share]
                new_row += [evaluated_value for evaluated_value in flow.get_all_evaluated_values()]
                rows.append(new_row)

        df = pd.DataFrame(rows, columns=col_names)
        return df

    def get_evaluated_flow_values_as_dataframe(self) -> DataFrame:
        """
        Get baseline evaluated value for Flows for all years.

        :return: DataFrame
        """

        unique_flows = self.get_unique_flows()
        sorted_flow_ids = sorted([flow.id for flow in unique_flows.values()], key=lambda x: x)
        columns = ["Year"]
        columns += [flow_id for flow_id in sorted_flow_ids]

        df = pd.DataFrame(columns=columns)
        df["Year"] = [year for year in self._years]
        df.set_index(["Year"], inplace=True)
        for year in df.index:
            for flow_id in df.columns:
                flow_value = 0.0
                if self.has_flow(flow_id, year):
                    flow = self.get_flow(flow_id, year)
                    if not isinstance(flow, Flow):
                        pass
                    else:
                        flow_value = flow.evaluated_value
                df.at[year, flow_id] = flow_value
        df.reset_index(inplace=True)
        return df

    def get_process(self, process_id: str, year: int = -1) -> Process:
        """
        Get Process by ID and target year.

        :param process_id: Process ID (str)
        :param year: Target year (int)
        :return: Process (Process)
        """
        if year >= 0:
            return self._year_to_process_id_to_process[year][process_id]

        return self._current_process_id_to_process[process_id]

    def get_flow(self, flow_id: str, year: int = -1) -> Flow:
        """
        Get Flow by ID and target year.

        :param flow_id: Flow ID (str)
        :param year: Target year (int)
        :return: Flow (Flow)
        """
        if year >= 0:
            return self._year_to_flow_id_to_flow[year][flow_id]

        return self._current_flow_id_to_flow[flow_id]

    def get_stock(self, process_id: str) -> Stock:
        """
        Get stock by ID.
        NOTE: Process and stocks share the same ID

        :param process_id: Target Stock ID / Process ID
        :return: Stock (Stock)
        """
        return self._process_id_to_stock[process_id]

    def get_baseline_dynamic_stocks(self) -> Dict[str, DynamicStockModel]:
        """
        Get dictionary of Stock ID -> baseline DynamicStockModel.

        :return: Dictionary (Stock ID -> baseline DynamicStockModel)
        """

        return self._stock_id_to_baseline_dsm

    def get_indicator_dynamic_stocks(self) -> Dict[str, Dict[str, DynamicStockModel]]:
        """
        Get dictionary of stock ID -> indicator name -> DynamicStockModel

        :return: Dictionary (stock ID (str) -> indicator name -> DynamicStockModel)
        """
        return self._stock_id_to_indicator_name_to_dsm

    def get_indicator_names(self) -> List[str]:
        """
        Get indicator names.

        :return: Indicator names (list of strings)
        """
        return list(self._indicator_name_to_indicator.keys())

    def get_indicator_name_to_indicator(self) -> Dict[str, Indicator]:
        """
        Get Indicator ID to Indicator mappings.

        :return: Dictionary (Indicator ID (str) -> Indicator (Indicator))
        """
        return self._indicator_name_to_indicator

    def get_process_inflows_total(self, process_id: str, year: int = -1) -> float:
        """
        Get total inflows (baseline) for Process ID.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: Sum of all inflows' evaluated value (baseline)
        """
        total = 0.0
        inflows = self._get_process_inflows(process_id, year)
        for flow in inflows:
            total += flow.evaluated_value
        return total

    def get_process_outflows_total(self, process_id: str, year: int = -1) -> float:
        """
        Get total outflows (baseline) for Process ID.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: Sum of all outflows' evaluated value (baseline)
        """
        total = 0.0
        outflows = self._get_process_outflows(process_id, year)
        for flow in outflows:
            total += flow.evaluated_value
        return total

    def get_process_outflows_total_abs(self, process_id: str, year: int = -1) -> float:
        """
        Get total absolute outflows evaluated value (baseline) for Process ID.
        Includes only absolute outflows.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: Sum of all absolute outflows' evaluated value (baseline)
        """
        total = 0.0
        outflows = self._get_process_outflows_abs(process_id, year)
        for flow in outflows:
            total += flow.evaluated_value
        return total

    def get_process_outflows_total_rel(self, process_id: str, year: int = -1) -> float:
        """
        Get total relative outflows evaluated value (baseline) for Process ID.
        Includes only relative outflows.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: Sum of all relative outflows' evaluated value (baseline)
        """
        total = 0.0
        outflows = self._get_process_outflows_rel(process_id, year)
        for flow in outflows:
            total += flow.evaluated_value
        return total

    def solve_timesteps(self) -> None:
        """
        Solves all timesteps.
        """
        bar = tqdm.tqdm(initial=0)
        self._create_dynamic_stocks()
        self._apply_flow_modifiers()
        self._remove_virtual_processes_and_flows()

        for current_year in self._years:
            bar.set_description("Solving flows for year {}/{}".format(current_year, self._year_end))
            self._solve_timestep()
            self._advance_timestep()
            bar.update()
        bar.close()

    def get_year_to_process_to_flows(self) -> Dict[int, Dict[Process, Dict[str, Flow]]]:
        """
        Get year to Process to Flow entry mappings

        :return: Year to Process to Flow entry mappings
        """
        year_to_process_to_flows = {}
        for year, process_id_to_process in self._year_to_process_id_to_process.items():
            year_to_process_to_flows[year] = {}
            for process_id, process in process_id_to_process.items():
                flow_ids = self._year_to_process_id_to_flow_ids[year][process_id]

                flows_in = []
                for flow_id in flow_ids["in"]:
                    flows_in.append(self._year_to_flow_id_to_flow[year][flow_id])

                flows_out = []
                for flow_id in flow_ids["out"]:
                    flows_out.append(self._year_to_flow_id_to_flow[year][flow_id])

                year_to_process_to_flows[year][process] = {"in": flows_in, "out": flows_out}

        return year_to_process_to_flows

    def get_process_flows(self, process_id: str, year: int = -1) -> Dict[str, List[Flow]]:
        """
        Get target Process ID flows for target year (both inflows and outflows).

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: Dictionary (keys: "Inflows", "Outflows"). Key points to List of Flows
        """
        process_inflows = self._get_process_inflows(process_id, year)
        process_outflows = self._get_process_outflows(process_id, year)
        return {"Inflows": process_inflows, "Outflows": process_outflows}

    def get_process_inflows(self, process_id: str, year: int = -1) -> List[Flow]:
        """
        Get target Process ID inflows.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: List of inflows
        """
        return self._get_process_inflows(process_id, year)

    def get_process_outflows(self, process_id: str, year: int = -1) -> List[Flow]:
        """
        Get target Process ID outflows.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: List of outflows
        """
        return self._get_process_outflows(process_id, year)

    def is_root_process(self, process_id: str, year: int = -1) -> bool:
        """
        Check if Process has no inflows at target year.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: True if Process has no inflows, False otherwise
        """
        return len(self._get_process_inflows(process_id, year)) == 0

    def is_leaf_process(self, process_id: str, year: int = -1) -> bool:
        """
        Check if Process ID has no outflows at target year.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: True if Process has no outflows, False otherwise
        """
        return len(self._get_process_outflows(process_id, year)) == 0

    def is_all_process_inflows_evaluated(self, process_id: str, year: int = -1) -> bool:
        """
        Check if all inflows to target Process ID are evaluated at the target year.

        :param process_id: Target Process ID (str)
        :param year: Target year (int)
        :return: True if all inflows are evaluated, False otherwise.
        """
        inflows = self._get_process_inflows(process_id, year)
        return all([flow.is_evaluated for flow in inflows])

    def has_flow(self, flow_id: str, year: int = -1) -> bool:
        """
        Check if Flow with ID exists at target year.
        If year is not provided then internally uses the current timestep year inside FlowSolver.

        :param flow_id: Flow ID (str)
        :param year: Target year (int)
        :return: True if Flow with ID exists for year, False otherwise.
        """

        if year >= 0:
            return flow_id in self._year_to_flow_id_to_flow[year]

        return flow_id in self._current_flow_id_to_flow

    def has_process(self, process_id: str, year: int = -1) -> bool:
        """
        Check if Process ID exists at target year.

        :param process_id: Process ID
        :param year: Selected year. If not defined then uses the current year inside FlowSolver.
        :return: True if Process with ID exists for year, False otherwise.
        """
        if year >= 0:
            return process_id in self._year_to_process_id_to_process[year]

        return process_id in self._current_process_id_to_process

    def accumulate_dynamic_stock_inflows(self, dsm: DynamicStockModel, total_inflows: float, year: int = -1) -> None:
        """
        Update and accumulate inflows to DynamicStockModel.

        :param dsm: Target DynamicStockModel
        :param total_inflows: Total inflows for the stock (float)
        :param year: Target year (int)
        :return: None
        """

        year_index = self._years.index(year)

        # Resetting some DynamicStockModel properties are needed to make
        # timestep stock accumulation and other calculations work
        dsm.i[year_index] = total_inflows

        # Recalculate stock by cohort
        dsm.s_c = None
        dsm.compute_s_c_inflow_driven()

        # Recalculate outflow by cohort
        dsm.o_c = None
        dsm.compute_o_c_from_s_c()

        # Recalculate stock total
        dsm.s = None
        dsm.compute_stock_total()

        # Get stock total
        dsm.compute_stock_change()

        # Recalculate stock outflow
        dsm.o = None
        dsm.compute_outflow_total()

    def get_year_range(self) -> List[int]:
        """
        Get years

        :return: List of years (integer)
        """
        return self._years

    def clamp_flow_values_below_zero(self) -> None:
        """
        Clamp all flows to minimum of 0.0 if flow value is negative.
        """
        # NOTE: Clamp all flows to minimum of 0.0 to introduce virtual flows
        for year, flow_id_to_flow in self._year_to_flow_id_to_flow.items():
            for flow_id, flow in flow_id_to_flow.items():
                if flow.value < 0.0:
                    flow.value = 0.0

                if flow.evaluated_value < 0.0:
                    flow.evaluated_value = 0.0

    def _get_year_to_process_id_to_process(self) -> Dict[int, Dict[str, Process]]:
        return self._year_to_process_id_to_process

    def _get_current_year_flow_id_to_flow(self) -> Dict[str, Flow]:
        """
        Get current year flow ID to Flow mappings.

        :return: Dictionary (Flow ID -> Flow)
        """
        return self._year_to_flow_id_to_flow[self._year_current]

    def _get_current_year_process_id_to_process(self) -> Dict[str, Process]:
        """
        Get current year process ID to Process mappings.

        :return: Dictionary (Process ID -> Process)
        """
        return self._year_to_process_id_to_process[self._year_current]

    def _get_current_year_process_id_to_to_flow_ids(self) -> Dict[str, Dict[str, List[str]]]:
        """
        Get current year Process ID to Flow ID mappings.

        :return: Dictionary (Process ID -> Dictionary (keys "in", "out") -> List of Flow IDs)
        """
        return self._year_to_process_id_to_flow_ids[self._year_current]

    def _get_year_to_flow_id_to_flow(self) -> Dict[int, Dict[str, Flow]]:
        """
        Get year to Flow ID to Flow mappings

        :return: Dictionary (Year -> Flow ID -> Flow)
        """
        return self._year_to_flow_id_to_flow

    def _get_process_inflow_ids(self, process_id: str, year: int = -1) -> List[str]:
        """
        Get list of inflow IDs to Process for target year.
        If year is not provided then uses the current year.

        :param process_id: Target Process ID
        :param year: Target year

        :return: List of inflow IDs (Flow)
        """
        result = []
        if year >= 0:
            result = self._year_to_process_id_to_flow_ids[year][process_id]["in"]
        else:
            result = self._current_process_id_to_flow_ids[process_id]["in"]

        # If year -> process ID does not exist, return empty array
        if not result:
            result = []

        return result

    def _get_process_outflow_ids(self, process_id: str, year: int = -1) -> List[str]:
        """
        Get list of outflow IDs from Process for target year.
        If year is not provided then uses the current year.

        :param process_id: Target Process ID
        :param year: Target year

        :return: List of outflow IDs (Flow)
        """

        result = []
        if year >= 0:
            result = self._year_to_process_id_to_flow_ids[year][process_id]["out"]
        else:
            result = self._current_process_id_to_flow_ids[process_id]["out"]

        if not result:
            result = []
        return result

    def _get_process_inflows(self, process_id: str, year: int = -1) -> List[Flow]:
        """
        Get list of all inflows for Process ID in target year.

        :param process_id: Target Process ID
        :param year: Target year
        :return: List of Flows
        """
        # Get list of process inflows for current year
        flows = []
        inflow_ids = self._get_process_inflow_ids(process_id, year)
        for flow_id in inflow_ids:
            flows.append(self.get_flow(flow_id, year))
        return flows

    # Get list of outflows (DataFlows)
    def _get_process_outflows(self, process_id: str, year: int = -1) -> List[Flow]:
        """
        Get list of Process outflows for the target year.

        :param process_id: Target Process ID
        :param year: Target year
        :return: List of Flows
        """
        # Get list of outflows for current year
        flows = []
        outflow_ids = self._get_process_outflow_ids(process_id, year)
        for flow_id in outflow_ids:
            flow = self.get_flow(flow_id, year)
            flows.append(flow)
        return flows

    def _get_process_outflows_abs(self, process_id: str, year: int = -1) -> List[Flow]:
        """
        Get list of absolute outflows from Process for the target year.

        :param process_id: Target Process ID
        :param year: Target year
        :return: List of absolute outflows (Flow)
        """
        outflows_abs = []
        flows = self._get_process_outflows(process_id, year)
        for flow in flows:
            if flow.is_unit_absolute_value:
                outflows_abs.append(flow)
        return outflows_abs

    def _get_process_outflows_rel(self, process_id: str, year: int = -1) -> List[Flow]:
        """
        Get list of relative outflows from Process for the target year.

        :param process_id: Target Process ID
        :param year: Target year
        :return: List of relative outflows (Flow)
        """

        outflows_rel = []
        flows = self._get_process_outflows(process_id, year)
        for flow in flows:
            if not flow.is_unit_absolute_value:
                outflows_rel.append(flow)
        return outflows_rel

    def _get_process_indicator_inflows_total(self, process_id: str, indicator_name: str, year: int = -1) -> float:
        """
        Get total evaluated inflows for Indicator for Process ID at target year.

        :param process_id: Target Process ID
        :param indicator_name: Target Indicator name
        :param year: Target year
        :return: Total inflows of indicator name (float)
        """
        total = 0.0
        flows = self._get_process_inflows(process_id, year)
        for flow in flows:
            total += flow.get_evaluated_value_for_indicator(indicator_name)
        return total

    def _get_process_indicator_outflows_total(self, process_id: str, indicator_name: str, year: int = -1) -> float:
        """
        Get total evaluated outflows for Indicator for Process ID at target year.

        :param process_id: Target Process ID
        :param indicator_name: Target Indicator name
        :param year: Target year
        :return: Total outflows of indicator name (float)
        """

        total = 0.0
        flows = self._get_process_outflows(process_id, year)
        for flow in flows:
            total += flow.get_evaluated_value_for_indicator(indicator_name)
        return total

    def _prepare_flows_for_timestep(self, flow_id_to_flow: Dict[str, Flow], year: int):
        """
        Prepare flows for timestep:
        - Mark all absolute flows as evaluated and set flow.value to flow.evaluated_value
        - Normalize all relative flow values from [0%, 100%] range to [0, 1] range
        - Mark all flows as prioritized that have target process in prioritized location or transformation stage

        :param flow_id_to_flow: Dictionary (Flow ID to Flow)
        """
        for flow_id, flow in flow_id_to_flow.items():
            if not isinstance(flow, Flow):
                continue

            if flow.is_unit_absolute_value:
                flow.is_evaluated = True
                flow.evaluated_share = 1.0
                flow.evaluated_value = flow.value
                flow.evaluate_indicator_values_from_baseline_value()
            else:
                # Normalize relative flow value from [0, 100] % range to 0 - 1 range
                if self._reset_evaluated_values:
                    flow.is_evaluated = False
                    flow.evaluated_share = flow.value / 100.0
                    flow.evaluated_value = 0.0
                else:
                    flow.is_evaluated = False
                    flow.evaluated_share = flow.value / 100.0

            # Mark flow prioritized
            process = self.get_process(flow.target_process_id, year)
            if process.location in self._prioritized_locations:
                flow.is_prioritized = True

            if process.transformation_stage in self._prioritized_transformation_stages:
                flow.is_prioritized = True

    def _evaluate_process(self, process_id: str, year: int) -> tuple[bool, List]:
        """
        Evaluate Process and accumulate inflows to DynamicStockModels.

        :param process_id: Target Process ID
        :param year: Target year
        :return: Tuple (is all inflows evaluated, list of outflows to check)
        """
        is_evaluated = False
        outflows = self._get_process_outflows(process_id, year)

        # Root process should have only absolute outflow
        if self.is_root_process(process_id, year):
            is_evaluated = True
            return is_evaluated, outflows

        # Distribute outflows (stock or direct) only if all the inflows are already evaluated
        if self.is_all_process_inflows_evaluated(process_id, year):
            # Total baseline inflows
            total_inflows = self.get_process_inflows_total(process_id, year)
            if process_id in self.get_baseline_dynamic_stocks():
                # All inflows are evaluated but process has stocks

                # Flow prioritization:
                # Ignore inflow amount to stock for outflows that are prioritized.
                total_outflows_prioritized = 0.0
                prioritized_outflows = {}
                for flow in outflows:
                    if not flow.is_prioritized:
                        continue

                    if not flow.is_unit_absolute_value:
                        raise Exception("Relative flow as prioritized flow!")

                    flow.is_evaluated = True
                    flow.evaluated_value = flow.value
                    prioritized_outflows[flow.id] = flow
                    total_outflows_prioritized += flow.evaluated_value

                if total_outflows_prioritized > total_inflows:
                    s = "Not enough inflows for prioritized outflows at process '{}' in year {}".format(
                        process_id, year)
                    sys.stdout.flush()
                    raise Exception(s)

                # Reduce total inflows to baseline stock by total prioritized outflows
                total_inflows_to_stock = total_inflows - total_outflows_prioritized

                # Update baseline DSM
                baseline_dsm = self.get_baseline_dynamic_stocks()[process_id]
                self.accumulate_dynamic_stock_inflows(baseline_dsm, total_inflows_to_stock, year)

                # Update stock inflows to indicator DSMs
                indicator_dynamic_stocks = self.get_indicator_dynamic_stocks()
                if process_id in indicator_dynamic_stocks:
                    inflows = self.get_process_inflows(process_id)
                    indicator_name_to_dsm = indicator_dynamic_stocks[process_id]
                    for indicator_name, indicator_dsm in indicator_name_to_dsm.items():
                        # Flow evaluated indicator values are based on the actual inflows to process with stock
                        # but with prioritized flows these values do not include the reduction of the prioritized flow.
                        # Fix this by...
                        # - calculating how much each inflow contributes to the original total inflows to get
                        #   correction factor)
                        # - Multiply total_inflows_to_stock by this factor to get correct value how much
                        #   flow indicator contributes to the indicator stock
                        total_indicator_inflows_to_stock = 0.0
                        if total_inflows_to_stock > 0.0:
                            for flow in inflows:
                                evaluated_indicator_value = flow.get_evaluated_value_for_indicator(indicator_name)
                                correction_factor = evaluated_indicator_value / total_inflows
                                corrected_flow_value = correction_factor * total_inflows_to_stock
                                total_indicator_inflows_to_stock += corrected_flow_value

                        self.accumulate_dynamic_stock_inflows(indicator_dsm, total_indicator_inflows_to_stock, year)

                # Distribute baseline total outflow values
                baseline_stock_outflow = self._get_dynamic_stock_outflow_value(baseline_dsm, year)

                # Check that if process has absolute outflow then outflow value must be
                # less than stock outflow. If absolute outflow is greater than stock outflow
                # then there is user error with the data
                outflows_abs = self._get_process_outflows_abs(process_id, year)
                outflows_rel = self._get_process_outflows_rel(process_id, year)

                # Get all outflows except prioritized outflows
                total_outflows_abs = np.sum([flow.evaluated_value for flow in outflows_abs if not flow.is_prioritized])
                total_outflows_rel = baseline_stock_outflow - total_outflows_abs
                if total_outflows_rel < 0.0:
                    # This is error: Total absolute outflows are greater than stock outflow.
                    # It means that there is not enough flows to distribute between remaining
                    # relative outflows
                    s = "Process {}: stock outflow ({:.3f}) is less than sum of absolute outflows ({:.3f}) in year {}!".format(
                        process_id, baseline_stock_outflow, total_outflows_rel, year)
                    raise Exception(s)

                # total_outflows_rel is the remaining outflows to be distributed between all relative outflows
                for flow in outflows_rel:
                    flow.is_evaluated = True
                    flow.evaluated_value = flow.evaluated_share * total_outflows_rel
                    flow.evaluate_indicator_values_from_baseline_value()

            else:
                # All inflows are evaluated but the current process does not have stocks
                outflows_abs = self._get_process_outflows_abs(process_id, year)
                outflows_rel = self._get_process_outflows_rel(process_id, year)
                total_outflows_abs = np.sum([flow.evaluated_value for flow in outflows_abs])

                # Ignore root and leaf processes because those have zero inflows and zero outflows
                is_root = self.is_root_process(process_id)
                is_leaf = self.is_leaf_process(process_id)

                # Check that virtual flows are actually needed
                diff = abs(total_inflows - total_outflows_abs)
                need_virtual_flows = total_inflows < total_outflows_abs and (diff > self._virtual_flows_epsilon)
                if (not is_root and not is_leaf) and total_inflows < total_outflows_abs:
                    if self._use_virtual_flows and need_virtual_flows:
                        # Create new virtual inflow and new virtual process where flow comes from
                        diff = total_inflows - total_outflows_abs
                        process = self.get_process(process_id, year)
                        v_process = self._create_virtual_process_ex(process)
                        v_flow = self._create_virtual_flow_ex(v_process, process, abs(diff))
                        v_flow.evaluate_indicator_values_from_baseline_value()

                        # Create virtual Flows and Processes to current year data
                        self._year_to_process_id_to_process[year][v_process.id] = v_process
                        self._year_to_process_id_to_flow_ids[year][v_process.id] = {"in": [], "out": []}
                        self._unique_process_id_to_process[v_process.id] = v_process

                        self._year_to_flow_id_to_flow[year][v_flow.id] = v_flow
                        self._year_to_process_id_to_flow_ids[year][v_flow.target_process_id]["in"].append(v_flow.id)
                        self._year_to_process_id_to_flow_ids[year][v_flow.source_process_id]["out"].append(v_flow.id)
                        self._unique_flow_id_to_flow[v_flow.id] = v_flow

                        # Recalculate total_inflows again
                        total_inflows = self.get_process_inflows_total(process_id)

                # Remaining outflows to be distributed between all relative outflows
                total_outflows_rel = total_inflows - total_outflows_abs
                for flow in outflows_rel:
                    flow.is_evaluated = True
                    flow.evaluated_value = flow.evaluated_share * total_outflows_rel
                    flow.evaluate_indicator_values_from_baseline_value()

            is_evaluated = True
            return is_evaluated, outflows

        for flow in outflows:
            if flow.is_unit_absolute_value:
                flow.is_evaluated = True

        return is_evaluated, outflows

    def _solve_timestep(self) -> None:
        """
        Solve current timestep.
        """
        self._current_flow_id_to_flow = self._year_to_flow_id_to_flow[self._year_current]
        self._current_process_id_to_flow_ids = self._year_to_process_id_to_flow_ids[self._year_current]
        self._current_process_id_to_process = self._year_to_process_id_to_process[self._year_current]

        # Mark all absolute flows as evaluated at the start of each timestep and also
        # mark all flows that have target process ID in prioritized transform stage as prioritized
        self._prepare_flows_for_timestep(self._current_flow_id_to_flow, self._year_current)

        # Each year evaluate dynamic stock outflows and related outflows as evaluated
        # NOTE: All outflows from process with stocks are initialized as evaluated relative flows
        # Without this mechanism the relative outflows from stocks are not possible, and it
        # would prevent in some cases the whole evaluation of scenarios with stocks.
        self._evaluate_dynamic_stock_outflows(self._year_current)

        # Add all root processes (= processes with no inflows) to unvisited list
        unevaluated_process_ids = []
        evaluated_process_ids = []
        current_year_process_ids = list(self._current_process_id_to_process.keys())
        for process_id in current_year_process_ids:
            inflows = self._get_process_inflows(process_id, year=self._year_current)
            if not inflows:
                unevaluated_process_ids.append(process_id)

        # Process flow value propagation until all inflows to processes are calculated
        current_iteration = 0
        while unevaluated_process_ids:
            process_id = unevaluated_process_ids.pop(0)
            if process_id in evaluated_process_ids:
                continue

            is_evaluated, outflows = self._evaluate_process(process_id, self._year_current)
            if is_evaluated:
                evaluated_process_ids.append(process_id)
                for flow in outflows:
                    target_process_id = flow.target_process_id
                    if target_process_id not in unevaluated_process_ids:
                        unevaluated_process_ids.insert(0, target_process_id)

            else:
                # Check all outflow target process ids
                for flow in outflows:
                    target_process_id = flow.target_process_id
                    if target_process_id not in unevaluated_process_ids:
                        unevaluated_process_ids.insert(0, target_process_id)

                # Add this process_id back to unevaluated list
                if process_id not in unevaluated_process_ids:
                    unevaluated_process_ids.append(process_id)

            # NOTE: Break out of infinite loop if running over big amount of iterations
            # This will happen if graph has loops that contain only relative flows between them
            current_iteration += 1
            if current_iteration >= self._max_iterations:
                print("Encountered processes that could not be evaluated in year {}:".format(self._year_current))
                print("The following processes have no inflows and have ONLY relative outflows (= error in data)")
                print("Possible ways to to fix:")
                print("- Introducing a valid inflow to the process")
                print("- Ensure that a valid inflow is present for the process in the model's initial year")
                print("")

                # Get list of unevaluated flows
                # The possible process causing the error is probably one of the flows' source processes
                unevaluated_inflows = []
                for p_id in current_year_process_ids:
                    for flow in self._get_process_inflows(p_id, self._year_current):
                        if not flow.is_evaluated:
                            unevaluated_inflows.append(flow)

                # Check all flow source processes and check for problematic processes:
                # Invalid process means:
                # - no inflows
                # - only relative outflows
                # This is definitely error in data
                unique_process_ids = set()
                for flow in unevaluated_inflows:
                    source_process_inflows = self._get_process_inflows(flow.source_process_id, self._year_current)
                    source_process_outflows = self._get_process_outflows(flow.source_process_id, self._year_current)
                    has_no_inflows = len(source_process_inflows) == 0
                    has_only_relative_outflows = len(source_process_outflows) > 0 and all(
                        [not flow.is_unit_absolute_value for flow in source_process_outflows])

                    if has_no_inflows and has_only_relative_outflows:
                        unique_process_ids.add(flow.source_process_id)

                print("List of invalid process IDs:")
                for source_process_id in unique_process_ids:
                    print("\t{}".format(source_process_id))
                print("")

                print("List of unevaluated flows:")
                for flow in unevaluated_inflows:
                    print("\t{}".format(flow))

                raise Exception("Unsolvable loop detected")

        # Check for unreported inflows or outflows (= process mass balance != 0)
        # and create virtual flows to balance out those processes.
        # Epsilon is maximum allowed difference of process inputs and outputs before creating virtual flow
        if self._use_virtual_flows:
            self._create_virtual_flows(self._year_current, self._virtual_flows_epsilon)

        # Show summary if virtual processes and flows created this year
        #self._show_virtual_flows_summary()

        # Recalculate evaluated values for stock outflows
        self._recalculate_indicator_dynamic_stock_outflows(self._year_current)

    def _advance_timestep(self) -> None:
        """
        Advance to next timestep.
        """
        self._year_prev = self._year_current
        self._year_current += 1

    def _create_virtual_process_id(self, process: Process) -> str:
        """
        Create virtual Process ID from target Process.

        :param process: Target Process
        :return: New vrtual Process ID
        """
        return self._virtual_process_id_prefix + process.id

    def _create_virtual_process_name(self, process: Process) -> str:
        """
        Create virtual Process name from target Process.

        :param process: Target Process
        :return: New virtual Process name
        """
        return self._virtual_process_id_prefix + process.name

    def _create_virtual_process_transformation_stage(self) -> str:
        """
        Get virtual process transformation stage.

        :return:
        """
        return self._virtual_process_transformation_stage

    def _create_virtual_process(self, process_id: str, process_name: str, transformation_stage: str) -> Process:
        """
        Create virtual Process

        :param process_id: Virtual Process ID
        :param process_name: Virtual Process name
        :param transformation_stage: Virtual process transformation stage
        :return: New virtual Process
        """
        new_virtual_process = Process()
        new_virtual_process.id = process_id
        new_virtual_process.name = process_name
        new_virtual_process.stock_lifetime = 0
        new_virtual_process.conversion_factor = 1.0
        new_virtual_process.transformation_stage = transformation_stage
        new_virtual_process.is_virtual = True
        return new_virtual_process

    def _create_virtual_process_ex(self, process: Process) -> Process:
        """
        Create virtual Process (extended).

        :param process: Target Process
        :return: New virtual Process
        """
        v_id = self._create_virtual_process_id(process)
        v_name = self._create_virtual_process_name(process)
        v_ts = self._create_virtual_process_transformation_stage()
        v_process = self._create_virtual_process(v_id, v_name, v_ts)
        return v_process

    def _remove_virtual_process(self, virtual_process_id: str, year: int) -> bool:
        """
        Remove process from year and all related virtual flows. Does nothing if process ID is not found.

        :param virtual_process_id: Target virtual process ID
        :param year: Target year
        :return: True if process ID was found, false otherwise
        """
        if year not in self._year_to_process_id_to_process:
            return False

        if virtual_process_id not in self._year_to_process_id_to_process[year]:
            return False

        # Remove flows related to this virtual process
        inflows = self.get_process_inflows(virtual_process_id, year)
        for flow in inflows:
            source_process_flow_ids = self._year_to_process_id_to_flow_ids[year][flow.source_process_id]["out"]
            if flow.id in source_process_flow_ids:
                source_process_flow_ids.remove(flow.id)
                if flow.id in self._year_to_flow_id_to_flow[year]:
                    del self._year_to_flow_id_to_flow[year][flow.id]
                    if flow.id in self._unique_flow_id_to_flow:
                        del self._unique_flow_id_to_flow[flow.id]

        outflows = self.get_process_outflows(virtual_process_id, year)
        for flow in outflows:
            target_process_inflow_ids = self._year_to_process_id_to_flow_ids[year][flow.target_process_id]["in"]
            if flow.id in target_process_inflow_ids:
                target_process_inflow_ids.remove(flow.id)
                if flow.id in self._year_to_flow_id_to_flow[year]:
                    del self._year_to_flow_id_to_flow[year][flow.id]
                    if flow.id in self._unique_flow_id_to_flow:
                        del self._unique_flow_id_to_flow[flow.id]

        # Remove virtual process
        if virtual_process_id in self._year_to_process_id_to_process[year]:
            del self._year_to_process_id_to_process[year][virtual_process_id]

        if virtual_process_id in self._year_to_process_id_to_flow_ids[year]:
            del self._year_to_process_id_to_flow_ids[year][virtual_process_id]

        if virtual_process_id in self._unique_process_id_to_process:
            del self._unique_process_id_to_process[virtual_process_id]

    def _create_virtual_flow(self, source_process_id: str, target_process_id: str, value: float, unit: str) -> Flow:
        """
        Create virtual flow.
        All indicator conversion factors are set to 0.0.

        :param source_process_id: Source Process ID
        :param target_process_id: Target Process ID
        :param value: Flow value
        :param unit: Flow unit (absolute/relative)
        :return: New virtual Flow
        """
        new_virtual_flow = Flow()
        new_virtual_flow.source_process_id = source_process_id
        new_virtual_flow.target_process_id = target_process_id
        new_virtual_flow.value = value
        new_virtual_flow.is_evaluated = True
        new_virtual_flow.evaluated_value = value
        new_virtual_flow.unit = unit
        new_virtual_flow.is_virtual = True

        # Copy indicators to virtual flows
        for indicator_name, indicator in self._indicators.items():
            new_indicator = copy.deepcopy(indicator)
            new_indicator.conversion_factor = 0.0
            new_virtual_flow.indicator_name_to_indicator[new_indicator.name] = new_indicator
            new_virtual_flow.indicator_name_to_evaluated_value[new_indicator.name] = 0.0

        return new_virtual_flow

    def _create_virtual_flow_ex(self, source_process: Process, target_process: Process, value: float) -> Flow:
        """
        Create virtual Flow (extended).

        :param source_process: Source Process
        :param target_process: Target Process
        :param value: Flow value
        :return: New virtual Flow
        """
        v_flow = self._create_virtual_flow(source_process.id, target_process.id, value, "")
        return v_flow

    def _create_virtual_flows(self, year: int, epsilon: float = 0.1) -> None:
        """
        Create virtual flows to balance out process inflows and outflows.
        NOTE: Virtual inflows are not created to processes with stocks.
        :param year: Target year
        :param epsilon: Maximum allowed absolute difference between total inflows and total outflows before creating
                        virtual flow
        """
        # Virtual outflow is unreported flow of process
        created_virtual_processes = {}
        created_virtual_flows = {}
        for process_id, process in self._current_process_id_to_process.items():
            # Skip virtual processes that were included during previous timesteps
            # to prevent cascading effect of creating infinite number of virtual processes and flows
            if process.is_virtual:
                continue

            # Ignore root and leaf processes (= root process has no inflows and leaf process has no outflows)
            is_root_process = self.is_root_process(process_id, year)
            is_leaf_process = self.is_leaf_process(process_id, year)
            if is_root_process or is_leaf_process:
                continue

            inflows_total = self.get_process_inflows_total(process_id, year)
            outflows_total = self.get_process_outflows_total(process_id, year)

            # If process has stock then consider only the stock outflows
            process_mass_balance = 0.0
            if process_id in self._process_id_to_stock:
                # Distribute baseline total outflow values
                baseline_dsm = self.get_baseline_dynamic_stocks()[process.id]
                baseline_stock_outflow = self._get_dynamic_stock_outflow_value(baseline_dsm, year)

                # Check that if process has absolute outflow then outflow value must be
                # less than stock outflow. If absolute outflow is greater than stock outflow
                # then there is user error with the data
                outflows_abs = self._get_process_outflows_abs(process_id, year)
                outflows_rel = self._get_process_outflows_rel(process_id, year)
                total_outflows_abs = np.sum([flow.evaluated_value for flow in outflows_abs if not flow.is_prioritized])
                total_outflows_rel = np.sum([flow.evaluated_value for flow in outflows_rel])
                process_mass_balance = baseline_stock_outflow - total_outflows_abs - total_outflows_rel
            else:
                # Process has no stock
                process_mass_balance = inflows_total - outflows_total

            if abs(process_mass_balance) < epsilon:
                # Total inflow and outflow difference less than epsilon, continue to next
                continue

            need_virtual_inflow = process_mass_balance < 0.0
            need_virtual_outflow = process_mass_balance > 0.0
            if not need_virtual_inflow and not need_virtual_outflow:
                # Inflows and outflows are balanced, do nothing and continue to next process
                continue

            #print("{}: Creating virtual flow, inflow={}, outflow={} ({})".format(process_id, need_virtual_inflow, need_virtual_outflow, year))

            if need_virtual_inflow:
                # Create new virtual Process
                v_process = self._create_virtual_process_ex(process)
                created_virtual_processes[v_process.id] = v_process

                # Create new virtual inflow
                source_process_id = v_process.id
                target_process_id = process_id
                value = process_mass_balance * -1.0
                unit = ""
                new_virtual_flow = self._create_virtual_flow(source_process_id, target_process_id, value, unit)
                created_virtual_flows[new_virtual_flow.id] = new_virtual_flow

            if need_virtual_outflow:
                # Create new virtual Process
                v_process = self._create_virtual_process_ex(process)
                created_virtual_processes[v_process.id] = v_process

                # Create new virtual outflow
                source_process_id = process_id
                target_process_id = v_process.id
                value = process_mass_balance
                unit = ""
                new_virtual_flow = self._create_virtual_flow(source_process_id, target_process_id, value, unit)
                created_virtual_flows[new_virtual_flow.id] = new_virtual_flow

        # Add create virtual Flows and Processes to current year data
        for v_id, virtual_process in created_virtual_processes.items():
            self._year_to_process_id_to_process[year][v_id] = virtual_process
            self._year_to_process_id_to_flow_ids[year][v_id] = {"in": [], "out": []}
            self._unique_process_id_to_process[v_id] = virtual_process

        for v_flow_id, virtual_flow in created_virtual_flows.items():
            self._year_to_flow_id_to_flow[year][v_flow_id] = virtual_flow
            self._year_to_process_id_to_flow_ids[year][virtual_flow.target_process_id]["in"].append(v_flow_id)
            self._year_to_process_id_to_flow_ids[year][virtual_flow.source_process_id]["out"].append(v_flow_id)
            self._unique_flow_id_to_flow[v_flow_id] = virtual_flow

    def _show_virtual_flows_summary(self):
        """
        Show virtual process and virtual flows summary for the current year.
        Does not do anything if no virtual processes or flows are created for current year.
        """
        # Show summary of created virtual processes and virtual flows
        virtual_processes = []
        virtual_flows = []
        for process_id, process in self._get_current_year_process_id_to_process().items():
            if not process.is_virtual:
                continue

            virtual_processes.append(process)

            entry = self._get_current_year_process_id_to_to_flow_ids()[process_id]
            inflow_ids = entry["in"]
            outflow_ids = entry["out"]

            for flow_id in inflow_ids:
                flow = self._get_current_year_flow_id_to_flow()[flow_id]
                if not flow.is_virtual:
                    continue

                virtual_flows.append(flow)

            for flow_id in outflow_ids:
                flow = self._get_current_year_flow_id_to_flow()[flow_id]
                if not flow.is_virtual:
                    continue

                virtual_flows.append(flow)

        # Show summary only if there is something to be shown
        num_virtual_processes = len(virtual_processes)
        num_virtual_flows = len(virtual_flows)
        if num_virtual_processes or num_virtual_flows:
            print("INFO: Created {} virtual processes and {} virtual flows for year {}".format(
                num_virtual_processes, num_virtual_flows, self._year_current))
            for process in virtual_processes:
                print("INFO: \t- Virtual process ID '{}'".format(process.id))
            for flow in virtual_flows:
                print("INFO: \t- Virtual flow ID '{} (value={:.5})'".format(flow.id, flow.evaluated_value))

            print("")

    def _create_dynamic_stocks(self) -> None:
        """
        Convert Stocks to ODYM DynamicStockModels.
        """
        # Create DynamicStockModels for Processes that contain Stock
        for stock in self.get_all_stocks():
            # If stock.distribution_params is float then use as default StdDev value
            # Otherwise check if the StdDev is defined for the cell
            stddev = 0.0
            shape = 1.0
            scale = 1.0
            if type(stock.stock_distribution_params) is float:
                stddev = stock.stock_distribution_params

            condition = None
            if type(stock.stock_distribution_params) is dict:
                stddev = stock.stock_distribution_params.get(StockDistributionParameter.StdDev, 1.0)
                shape = stock.stock_distribution_params.get(StockDistributionParameter.Shape, 1.0)
                scale = stock.stock_distribution_params.get(StockDistributionParameter.Scale, 1.0)

                # For new decay functions
                landfill_decay_types = [StockDistributionType.LandfillDecayWood, StockDistributionType.LandfillDecayWood]
                if stock.stock_distribution_type in landfill_decay_types:
                    condition = stock.stock_distribution_params[StockDistributionParameter.Condition]

            # Stock parameters
            stock_years = np.array(self._years)
            stock_total_inflows = np.zeros(len(stock_years))
            stock_total = np.zeros(len(stock_years))
            stock_lifetime_params = {
                'Type': stock.stock_distribution_type,
                'Mean': [stock.stock_lifetime],
                'StdDev': [stddev],
                'Shape': [shape],
                'Scale': [scale],
                StockDistributionParameter.Condition: [condition],
            }

            # Baseline DSM
            baseline_dsm = DynamicStockModel(t=copy.deepcopy(stock_years),
                                             i=copy.deepcopy(stock_total_inflows),
                                             s=copy.deepcopy(stock_total),
                                             lt=copy.deepcopy(stock_lifetime_params))

            baseline_dsm.compute_s_c_inflow_driven()
            baseline_dsm.compute_o_c_from_s_c()
            baseline_dsm.compute_stock_total()
            baseline_dsm.compute_stock_change()
            baseline_dsm.compute_outflow_total()

            # Stock ID -> DSM
            self._stock_id_to_baseline_dsm[stock.id] = baseline_dsm

            # Create indicator DSMs for each indicator name
            for indicator_name in self.get_indicator_names():
                indicator_dsm = DynamicStockModel(t=copy.deepcopy(stock_years),
                                                  i=copy.deepcopy(stock_total_inflows),
                                                  s=copy.deepcopy(stock_total),
                                                  lt=copy.deepcopy(stock_lifetime_params))

                indicator_dsm.compute_s_c_inflow_driven()
                indicator_dsm.compute_o_c_from_s_c()
                indicator_dsm.compute_stock_total()
                indicator_dsm.compute_stock_change()
                indicator_dsm.compute_outflow_total()

                # Stock ID -> Indicator name -> DSM
                indicator_name_to_dsm = self._stock_id_to_indicator_name_to_dsm.get(stock.id, {})
                indicator_name_to_dsm[indicator_name] = indicator_dsm
                self._stock_id_to_indicator_name_to_dsm[stock.id] = indicator_name_to_dsm

    def _evaluate_dynamic_stock_outflows(self, year: int) -> None:
        """
        Evaluate dynamic stock outflows and distribute stock outflow among all outflows.
        Marks stock outflows as evaluated.

        This method must be called at the beginning of every timestep before starting evaluating Processes.

        :param year: Target year
        """
        # Get stock outflow for year, distribute that to relative outflows and mark those Flows as evaluated
        # NOTE: Now also outflows to prioritized flows
        year_index = self._years.index(year)
        for stock_id, dsm in self.get_baseline_dynamic_stocks().items():
            outflows = self._get_process_outflows(stock_id)
            stock_total_outflow = dsm.compute_outflow_total()[year_index]

            for flow in outflows:
                if flow.is_unit_absolute_value:
                    continue

                flow.is_evaluated = True
                flow.evaluated_value = flow.evaluated_share * stock_total_outflow
                flow.evaluate_indicator_values_from_baseline_value()

    def _recalculate_indicator_dynamic_stock_outflows(self, year: int) -> None:
        # Recalculate indicator values for stock outflows
        # NOTE: Relative outflows from indicator stock do not need indicator conversion factor
        # defined in the settings file because relative share of that flow can be directly calculated
        # from indicator stock total outflow. Indicator conversion factors are needed when flows enter stock
        # but in stock outflow it's not mandatory.
        year_index = self._years.index(year)
        for stock_id, dsm_indicators in self.get_indicator_dynamic_stocks().items():
            outflows = self._get_process_outflows(stock_id)
            for indicator_name, dsm in dsm_indicators.items():
                stock_total_outflow = dsm.compute_outflow_total()[year_index]
                for flow in outflows:
                    if flow.is_unit_absolute_value:
                        continue

                    evaluated_value = flow.evaluated_share * stock_total_outflow
                    flow.set_evaluated_value_for_indicator(indicator_name, evaluated_value)

    def _get_dynamic_stock_outflow_value(self, dsm: DynamicStockModel, year: int) -> float:
        """
        Get dynamic stock total outflow value.

        :param dsm: Target DynamicStockModel
        :param year: Target year
        :return: Total stock outflow (float)
        """
        year_index = self._years.index(year)
        stock_outflow_total = dsm.compute_outflow_total()
        return stock_outflow_total[year_index]

    def get_solved_scenario_data(self) -> ScenarioData:
        """
        Get solved ScenarioData.

        :return: Solved ScenarioData
        """
        # Make deep copies and return ScenarioData containing the data
        years = copy.deepcopy(self._years)
        year_to_process_id_to_process = copy.deepcopy(self._year_to_process_id_to_process)
        year_to_process_id_to_flow_ids = copy.deepcopy(self._year_to_process_id_to_flow_ids)
        year_to_flow_id_to_flow = copy.deepcopy(self._year_to_flow_id_to_flow)
        unique_process_id_to_process = copy.deepcopy(self._unique_process_id_to_process)
        unique_flow_id_to_flow = copy.deepcopy(self._unique_flow_id_to_flow)
        process_id_to_stock = copy.deepcopy(self._process_id_to_stock)
        stocks = copy.deepcopy(self._all_stocks)
        use_virtual_flows = copy.deepcopy(self._use_virtual_flows)
        virtual_flows_epsilon = copy.deepcopy(self._virtual_flows_epsilon)
        baseline_value_name = copy.deepcopy(self._baseline_value_name)
        baseline_unit_name = copy.deepcopy(self._baseline_unit_name)
        indicator_name_to_indicator = copy.deepcopy(self._indicator_name_to_indicator)

        scenario_data = ScenarioData(years=years,
                                     year_to_process_id_to_process=year_to_process_id_to_process,
                                     year_to_process_id_to_flow_ids=year_to_process_id_to_flow_ids,
                                     year_to_flow_id_to_flow=year_to_flow_id_to_flow,
                                     unique_process_id_to_process=unique_process_id_to_process,
                                     unique_flow_id_to_flow=unique_flow_id_to_flow,
                                     process_id_to_stock=process_id_to_stock,
                                     stocks=stocks,
                                     use_virtual_flows=use_virtual_flows,
                                     virtual_flows_epsilon=virtual_flows_epsilon,
                                     baseline_value_name=baseline_value_name,
                                     baseline_unit_name=baseline_unit_name,
                                     indicator_name_to_indicator=indicator_name_to_indicator
        )
        return scenario_data

    def _apply_flow_modifiers(self) -> None:
        """
        Apply flow modifiers if Scenario has those defined.
        Baseline scenario does not have those so return immediately if there isn't anything to process.
        """
        # NOTE: Lazily import FlowModifierSolver to fix circular dependency
        from .flowmodifiersolver import FlowModifierSolver

        if not self._scenario.scenario_definition.flow_modifiers:
            # This is the case when dealing with baseline scenario, do nothing and just return
            return

        print("*** Applying flow modifiers for scenario '{}' ***".format(self._scenario.name))
        scenario_type = self._scenario.model_params[ParameterName.ScenarioType]
        fms = FlowModifierSolver(self, scenario_type)
        fms.solve()

    def _remove_virtual_processes_and_flows(self) -> None:
        # Remove all virtual processes and related flows in all years
        for year in self._years:
            virtual_process_ids = []
            for process_id, process in self._year_to_process_id_to_process[year].items():
                if process.is_virtual:
                    virtual_process_ids.append(process_id)

            for process_id in virtual_process_ids:
                self._remove_virtual_process(process_id, year)
